@@ -26,11 +26,13 @@
 #include "memory.h"
 #include "syntax.h"
 #include "transform.h"
+#include "errcode.h"
 
 #include <fnmatch.h>
 #include <argz.h>
 #include <string.h>
 #include <stdarg.h>
+#include <locale.h>
 
 /* Some popular labels that we use in /augeas */
 static const char *const s_augeas = "augeas";
@@ -51,11 +53,24 @@ static const char *const static_nodes[][2] = {
     { AUGEAS_META_TREE "/version/defvar", NULL }
 };
 
+static const char *const errcodes[] = {
+    "No error",                                         /* AUG_NOERROR */
+    "Cannot allocate memory",                           /* AUG_ENOMEM */
+    "Internal error (please file a bug)",               /* AUG_EINTERNAL */
+    "Invalid path expression",                          /* AUG_EPATHX */
+    "No match for path expression",                     /* AUG_ENOMATCH */
+    "Too many matches for path expression",             /* AUG_EMMATCH */
+    "Syntax error in lens definition",                  /* AUG_ESYNTAX */
+    "Lens not found",                                   /* AUG_ENOLENS */
+    "Multiple transforms"                               /* AUG_EMXFM */
+};
+
 static void tree_mark_dirty(struct tree *tree) {
     do {
         tree->dirty = 1;
         tree = tree->parent;
     } while (tree != tree->parent && !tree->dirty);
+    tree->dirty = 1;
 }
 
 /* Clear the dirty flag in the whole TREE */
@@ -67,7 +82,7 @@ static void tree_clean(struct tree *tree) {
     tree->dirty = 0;
 }
 
-static struct tree *tree_child(struct tree *tree, const char *label) {
+struct tree *tree_child(struct tree *tree, const char *label) {
     if (tree == NULL)
         return NULL;
 
@@ -78,8 +93,7 @@ static struct tree *tree_child(struct tree *tree, const char *label) {
     return NULL;
 }
 
-/* Get first existing child or create one */
-static struct tree *tree_child_cr(struct tree *tree, const char *label) {
+struct tree *tree_child_cr(struct tree *tree, const char *label) {
     static struct tree *child = NULL;
 
     if (tree == NULL)
@@ -95,7 +109,7 @@ static struct tree *tree_child_cr(struct tree *tree, const char *label) {
     return child;
 }
 
-static struct tree *tree_path_cr(struct tree *tree, int n, ...) {
+struct tree *tree_path_cr(struct tree *tree, int n, ...) {
     va_list ap;
 
     va_start(ap, n);
@@ -107,7 +121,42 @@ static struct tree *tree_path_cr(struct tree *tree, int n, ...) {
     return tree;
 }
 
-static int tree_set_value(struct tree *tree, const char *value) {
+struct tree *tree_find(struct augeas *aug, const char *path) {
+    struct pathx *p = NULL;
+    struct tree *result = NULL;
+    int r;
+
+    pathx_parse(aug->origin, aug->error, path, true, aug->symtab, &p);
+    ERR_BAIL(aug);
+
+    r = pathx_find_one(p, &result);
+    BUG_ON(r > 1, aug,
+           "Multiple matches for %s when only one was expected",
+           path);
+ done:
+    free_pathx(p);
+    return result;
+ error:
+    result = NULL;
+    goto done;
+}
+
+struct tree *tree_find_cr(struct augeas *aug, const char *path) {
+    struct pathx *p = NULL;
+    struct tree *result = NULL;
+    int r;
+
+    pathx_parse(aug->origin, aug->error, path, true, aug->symtab, &p);
+    ERR_BAIL(aug);
+
+    r = pathx_expand_tree(p, &result);
+    ERR_BAIL(aug);
+ error:
+    free_pathx(p);
+    return result;
+}
+
+int tree_set_value(struct tree *tree, const char *value) {
     if (tree->value != NULL) {
         free(tree->value);
         tree->value = NULL;
@@ -121,34 +170,32 @@ static int tree_set_value(struct tree *tree, const char *value) {
     return 0;
 }
 
-/* Parse a path expression. Report errors in /augeas/pathx/error */
-static struct pathx *parse_user_pathx(const struct augeas *aug,
-                                      bool need_nodeset,
-                                      const char *path) {
-    struct pathx *result;
-    int    pos;
-
-    if (pathx_parse(aug->origin, path, need_nodeset, aug->symtab, &result)
-        == PATHX_NOERROR)
-        return result;
+/* Report pathx errors in /augeas/pathx/error */
+static void store_pathx_error(const struct augeas *aug) {
+    if (aug->error->code != AUG_EPATHX)
+        return;
 
     struct tree *error =
         tree_path_cr(aug->origin, 3, s_augeas, s_pathx, s_error);
     if (error == NULL)
-        return NULL;
-    tree_set_value(error, pathx_error(result, NULL, &pos));
+        return;
+    tree_set_value(error, aug->error->minor_details);
 
     struct tree *tpos = tree_child_cr(error, s_pos);
     if (tpos == NULL)
-        return NULL;
+        return;
+    tree_set_value(tpos, aug->error->details);
+}
 
-    free(tpos->value);
-    if (ALLOC_N(tpos->value, strlen(path) + 4) >= 0) {
-        strncpy(tpos->value, path, pos);
-        strcat(tpos->value, "|=|");
-        strcat(tpos->value, path + pos);
-    }
-    return NULL;
+static struct pathx *parse_user_pathx(const struct augeas *aug,
+                                      bool need_nodeset,
+                                      const char *path) {
+    struct pathx *result;
+    struct error *err = err_of_aug(aug);
+
+    pathx_parse(aug->origin, err, path, need_nodeset, aug->symtab,
+                &result);
+    return result;
 }
 
 static const char *init_root(const char *root0) {
@@ -156,7 +203,7 @@ static const char *init_root(const char *root0) {
 
     if (root0 == NULL)
         root0 = getenv(AUGEAS_ROOT_ENV);
-    if (root0 == NULL)
+    if (root0 == NULL || root0[0] == '\0')
         root0 = "/";
     root = strdup(root0);
     if (root == NULL)
@@ -219,6 +266,61 @@ static struct tree *tree_from_transform(struct augeas *aug,
     return NULL;
 }
 
+/* Save user locale and switch to C locale */
+#if HAVE_USELOCALE
+static void save_locale(struct augeas *aug) {
+    if (aug->c_locale == NULL) {
+        aug->c_locale = newlocale(LC_ALL_MASK, "C", NULL);
+        ERR_NOMEM(aug->c_locale == NULL, aug);
+    }
+
+    aug->user_locale = uselocale(aug->c_locale);
+ error:
+    return;
+}
+#else
+static void save_locale(ATTRIBUTE_UNUSED struct augeas *aug) { }
+#endif
+
+#if HAVE_USELOCALE
+static void restore_locale(struct augeas *aug) {
+    uselocale(aug->user_locale);
+    aug->user_locale = NULL;
+}
+#else
+static void restore_locale(ATTRIBUTE_UNUSED struct augeas *aug) { }
+#endif
+
+/* Clean up old error messages every time we enter through the public
+ * API. Since we make internal calls through the public API, we keep a
+ * count of how many times a public API call was made, and only reset when
+ * that count is 0. That requires that all public functions enclose their
+ * work within a matching pair of api_entry/api_exit calls.
+ */
+static void api_entry(const struct augeas *aug) {
+    struct error *err = ((struct augeas *) aug)->error;
+
+    ((struct augeas *) aug)->api_entries += 1;
+
+    if (aug->api_entries > 1)
+        return;
+
+    err->code = AUG_NOERROR;
+    err->minor = 0;
+    FREE(err->details);
+    err->minor_details = NULL;
+    save_locale((struct augeas *) aug);
+}
+
+static void api_exit(const struct augeas *aug) {
+    assert(aug->api_entries > 0);
+    ((struct augeas *) aug)->api_entries -= 1;
+    if (aug->api_entries == 0) {
+        store_pathx_error(aug);
+        restore_locale((struct augeas *) aug);
+    }
+}
+
 struct augeas *aug_init(const char *root, const char *loadpath,
                         unsigned int flags) {
     struct augeas *result;
@@ -229,12 +331,18 @@ struct augeas *aug_init(const char *root, const char *loadpath,
 
     if (ALLOC(result) < 0)
         goto error;
+    if (ALLOC(result->error) < 0)
+        goto error;
+    result->error->aug = result;
+
     result->origin = make_tree_origin(tree_root);
     if (result->origin == NULL) {
         free_tree(tree_root);
         goto error;
     }
 
+    api_entry(result);
+
     result->flags = flags;
 
     result->root = init_root(root);
@@ -259,7 +367,8 @@ struct augeas *aug_init(const char *root, const char *loadpath,
     if (result->nmodpath > 0) {
         argz_stringify(result->modpathz, result->nmodpath, PATH_SEP_CHAR);
         char *s, *t;
-        for (s = result->modpathz, t = result->modpathz; *s != '\0'; s++) {
+        const char *e = result->modpathz + strlen(result->modpathz);
+        for (s = result->modpathz, t = result->modpathz; s < e; s++) {
             char *p = s;
             if (*p == '/') {
                 while (*p == '/') p += 1;
@@ -312,17 +421,21 @@ struct augeas *aug_init(const char *root, const char *loadpath,
         if (aug_load(result) < 0)
             goto error;
 
+    api_exit(result);
     return result;
 
  error:
+    api_exit(result);
     aug_close(result);
     return NULL;
 }
 
-static void tree_unlink_children(struct tree *tree) {
+void tree_unlink_children(struct augeas *aug, struct tree *tree) {
     if (tree == NULL)
         return;
 
+    pathx_symtab_remove_descendants(aug->symtab, tree);
+
     while (tree->children != NULL)
         tree_unlink(tree->children);
 }
@@ -333,18 +446,41 @@ int aug_load(struct augeas *aug) {
     struct tree *files = tree_child_cr(aug->origin, s_files);
     struct tree *load = tree_child_cr(meta, s_load);
 
-    if (load == NULL)
-        return -1;
+    api_entry(aug);
 
-    tree_unlink_children(meta_files);
-    tree_unlink_children(files);
+    ERR_NOMEM(load == NULL, aug);
+
+    tree_unlink_children(aug, meta_files);
+    tree_unlink_children(aug, files);
 
     list_for_each(xfm, load->children) {
         if (transform_validate(aug, xfm) == 0)
             transform_load(aug, xfm);
     }
     tree_clean(aug->origin);
+
+    api_exit(aug);
     return 0;
+ error:
+    api_exit(aug);
+    return -1;
+}
+
+static int find_one_node(struct pathx *p, struct tree **match) {
+    struct error *err = err_of_pathx(p);
+    int r = pathx_find_one(p, match);
+
+    if (r == 1)
+        return 0;
+
+    if (r == 0) {
+        report_error(err, AUG_ENOMATCH, NULL);
+    } else {
+        /* r > 1 */
+        report_error(err, AUG_EMMATCH, NULL);
+    }
+
+    return -1;
 }
 
 int aug_get(const struct augeas *aug, const char *path, const char **value) {
@@ -352,9 +488,10 @@ int aug_get(const struct augeas *aug, const char *path, const char **value) {
     struct tree *match;
     int r;
 
+    api_entry(aug);
+
     p = parse_user_pathx((struct augeas *) aug, true, path);
-    if (p == NULL)
-        return -1;
+    ERR_BAIL(aug);
 
     if (value != NULL)
         *value = NULL;
@@ -364,23 +501,29 @@ int aug_get(const struct augeas *aug, const char *path, const char **value) {
         *value = match->value;
     free_pathx(p);
 
+    api_exit(aug);
     return r;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 int aug_defvar(augeas *aug, const char *name, const char *expr) {
     struct pathx *p = NULL;
     int result = -1;
 
+    api_entry(aug);
+
     if (expr == NULL) {
         result = pathx_symtab_undefine(&(aug->symtab), name);
     } else {
         p = parse_user_pathx((struct augeas *) aug, false, expr);
-        if (p == NULL)
-            goto done;
+        ERR_BAIL(aug);
         result = pathx_symtab_define(&(aug->symtab), name, p);
     }
- done:
+ error:
     free_pathx(p);
+    api_exit(aug);
     return result;
 }
 
@@ -391,14 +534,15 @@ int aug_defnode(augeas *aug, const char *name, const char *expr,
     int r, cr;
     struct tree *tree;
 
+    api_entry(aug);
+
     if (expr == NULL)
-        return -1;
+        goto error;
     if (created == NULL)
         created = &cr;
 
     p = parse_user_pathx((struct augeas *) aug, false, expr);
-    if (p == NULL)
-        goto done;
+    ERR_BAIL(aug);
 
     r = pathx_expand_tree(p, &tree);
     if (r < 0)
@@ -409,13 +553,18 @@ int aug_defnode(augeas *aug, const char *name, const char *expr,
         r = tree_set_value(tree, value);
         if (r < 0)
             goto done;
+        result = pathx_symtab_assign_tree(&(aug->symtab), name, tree);
+    } else {
+        result = pathx_symtab_define(&(aug->symtab), name, p);
     }
 
-    result = pathx_symtab_define(&(aug->symtab), name, p);
-
  done:
     free_pathx(p);
+    api_exit(aug);
     return result;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 struct tree *tree_set(struct pathx *p, const char *value) {
@@ -436,13 +585,19 @@ int aug_set(struct augeas *aug, const char *path, const char *value) {
     struct pathx *p;
     int result;
 
+    api_entry(aug);
+
     p = parse_user_pathx(aug, true, path);
-    if (p == NULL)
-        return -1;
+    ERR_BAIL(aug);
 
     result = tree_set(p, value) == NULL ? -1 : 0;
     free_pathx(p);
+
+    api_exit(aug);
     return result;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 int tree_insert(struct pathx *p, const char *label, int before) {
@@ -451,7 +606,7 @@ int tree_insert(struct pathx *p, const char *label, int before) {
     if (strchr(label, SEP) != NULL)
         return -1;
 
-    if (pathx_find_one(p, &match) != 1)
+    if (find_one_node(p, &match) < 0)
         goto error;
 
     new = make_tree(strdup(label), NULL, match->parent, NULL);
@@ -475,13 +630,15 @@ int aug_insert(struct augeas *aug, const char *path, const char *label,
     struct pathx *p = NULL;
     int result = -1;
 
+    api_entry(aug);
+
     p = parse_user_pathx(aug, true, path);
-    if (p == NULL)
-        goto done;
+    ERR_BAIL(aug);
 
     result = tree_insert(p, label, before);
- done:
+ error:
     free_pathx(p);
+    api_exit(aug);
     return result;
 }
 
@@ -571,7 +728,7 @@ int tree_rm(struct pathx *p) {
     for (i = 0, tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (TREE_HIDDEN(tree))
             continue;
-        pathx_symtab_remove_descendants(p, tree);
+        pathx_symtab_remove_descendants(pathx_get_symtab(p), tree);
         del[i] = tree;
         i += 1;
     }
@@ -587,14 +744,20 @@ int aug_rm(struct augeas *aug, const char *path) {
     struct pathx *p = NULL;
     int result;
 
+    api_entry(aug);
+
     p = parse_user_pathx(aug, true, path);
-    if (p == NULL)
-        return -1;
+    ERR_BAIL(aug);
 
     result = tree_rm(p);
     free_pathx(p);
+    ERR_BAIL(aug);
 
+    api_exit(aug);
     return result;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 int tree_replace(struct tree *origin, const char *path, struct tree *sub) {
@@ -602,7 +765,7 @@ int tree_replace(struct tree *origin, const char *path, struct tree *sub) {
     struct pathx *p = NULL;
     int r;
 
-    if (pathx_parse(origin, path, true, NULL, &p) != PATHX_NOERROR)
+    if (pathx_parse(origin, NULL, path, true, NULL, &p) != PATHX_NOERROR)
         goto error;
 
     r = tree_rm(p);
@@ -631,28 +794,28 @@ int aug_mv(struct augeas *aug, const char *src, const char *dst) {
     struct tree *ts, *td, *t;
     int r, ret;
 
+    api_entry(aug);
+
     ret = -1;
     s = parse_user_pathx(aug, true, src);
-    if (s == NULL)
-        goto done;
+    ERR_BAIL(aug);
 
     d = parse_user_pathx(aug, true, dst);
-    if (d == NULL)
-        goto done;
+    ERR_BAIL(aug);
 
-    r = pathx_find_one(s, &ts);
-    if (r != 1)
-        goto done;
+    r = find_one_node(s, &ts);
+    if (r < 0)
+        goto error;
 
     r = pathx_expand_tree(d, &td);
     if (r == -1)
-        goto done;
+        goto error;
 
     /* Don't move SRC into its own descendent */
     t = td;
     do {
         if (t == ts)
-            goto done;
+            goto error;
         t = t->parent;
     } while (! ROOT_P(t));
 
@@ -672,9 +835,10 @@ int aug_mv(struct augeas *aug, const char *src, const char *dst) {
     tree_mark_dirty(td);
 
     ret = 0;
- done:
+ error:
     free_pathx(s);
     free_pathx(d);
+    api_exit(aug);
     return ret;
 }
 
@@ -683,6 +847,8 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
     struct tree *tree;
     int cnt = 0;
 
+    api_entry(aug);
+
     if (matches != NULL)
         *matches = NULL;
 
@@ -691,13 +857,13 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
     }
 
     p = parse_user_pathx((struct augeas *) aug, true, pathin);
-    if (p == NULL)
-        return -1;
+    ERR_BAIL(aug);
 
     for (tree = pathx_first(p); tree != NULL; tree = pathx_next(p)) {
         if (! TREE_HIDDEN(tree))
             cnt += 1;
     }
+    ERR_BAIL(aug);
 
     if (matches == NULL)
         goto done;
@@ -715,8 +881,10 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
         }
         i += 1;
     }
+    ERR_BAIL(aug);
  done:
     free_pathx(p);
+    api_exit(aug);
     return cnt;
 
  error:
@@ -728,6 +896,7 @@ int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
         }
     }
     free_pathx(p);
+    api_exit(aug);
     return -1;
 }
 
@@ -755,7 +924,17 @@ static int tree_save(struct augeas *aug, struct tree *tree,
                     if (transform == NULL || transform == xfm) {
                         transform = xfm;
                     } else {
-                        FIXME("Multiple transforms for %s", path);
+                        const char *filename =
+                            tpath + strlen(AUGEAS_FILES_TREE) + 1;
+                        transform_file_error(aug, "mxfm_save", filename,
+                           "Lenses %s and %s could be used to save this file",
+                                             xfm_lens_name(transform),
+                                             xfm_lens_name(xfm));
+                        ERR_REPORT(aug, AUG_EMXFM,
+                                   "Path %s transformable by lens %s and %s",
+                                   tpath,
+                                   xfm_lens_name(transform),
+                                   xfm_lens_name(xfm));
                         result = -1;
                     }
                 }
@@ -818,7 +997,7 @@ static int unlink_removed_files(struct augeas *aug,
         if (tf == NULL) {
             /* Unlink all files in tm */
             struct pathx *px = NULL;
-            if (pathx_parse(tm, file_nodes, true, NULL, &px)
+            if (pathx_parse(tm, err_of_aug(aug), file_nodes, true, NULL, &px)
                 != PATHX_NOERROR) {
                 result = -1;
                 continue;
@@ -845,11 +1024,13 @@ int aug_save(struct augeas *aug) {
     struct tree *files = tree_child_cr(aug->origin, s_files);
     struct tree *load = tree_child_cr(meta, s_load);
 
+    api_entry(aug);
+
     if (update_save_flags(aug) < 0)
-        return -1;
+        goto error;
 
     if (files == NULL || meta == NULL || load == NULL)
-        return -1;
+        goto error;
 
     aug_rm(aug, AUGEAS_EVENTS_SAVED);
 
@@ -870,7 +1051,12 @@ int aug_save(struct augeas *aug) {
     if (!(aug->flags & AUG_SAVE_NOOP)) {
         tree_clean(aug->origin);
     }
+
+    api_exit(aug);
     return ret;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 static int print_one(FILE *out, const char *path, const char *value) {
@@ -952,7 +1138,7 @@ int dump_tree(FILE *out, struct tree *tree) {
     struct pathx *p;
     int result;
 
-    if (pathx_parse(tree, "/*", true, NULL, &p) != PATHX_NOERROR)
+    if (pathx_parse(tree, NULL, "/*", true, NULL, &p) != PATHX_NOERROR)
         return -1;
 
     result = print_tree(out, p, 1);
@@ -964,31 +1150,47 @@ int aug_print(const struct augeas *aug, FILE *out, const char *pathin) {
     struct pathx *p;
     int result;
 
+    api_entry(aug);
+
     if (pathin == NULL || strlen(pathin) == 0) {
         pathin = "/*";
     }
 
     p = parse_user_pathx((struct augeas *) aug, true, pathin);
-    if (p == NULL)
-        return -1;
+    ERR_BAIL(aug);
 
     result = print_tree(out, p, 0);
     free_pathx(p);
 
+    api_exit(aug);
     return result;
+ error:
+    api_exit(aug);
+    return -1;
 }
 
 void aug_close(struct augeas *aug) {
     if (aug == NULL)
         return;
+
+    /* There's no point in bothering with api_entry/api_exit here */
     free_tree(aug->origin);
     unref(aug->modules, module);
     free((void *) aug->root);
     free(aug->modpathz);
     free_symtab(aug->symtab);
+    free(aug->error->details);
+    free(aug->error);
     free(aug);
 }
 
+int __aug_load_module_file(struct augeas *aug, const char *filename) {
+    api_entry(aug);
+    int r = load_module_file(aug, filename);
+    api_exit(aug);
+    return r;
+}
+
 int tree_equal(const struct tree *t1, const struct tree *t2) {
     while (t1 != NULL && t2 != NULL) {
         if (!streqv(t1->label, t2->label))
@@ -1003,6 +1205,28 @@ int tree_equal(const struct tree *t1, const struct tree *t2) {
     return t1 == t2;
 }
 
+/*
+ * Error reporting API
+ */
+int aug_error(struct augeas *aug) {
+    return aug->error->code;
+}
+
+const char *aug_error_message(struct augeas *aug) {
+    aug_errcode_t errcode = aug->error->code;
+
+    if (errcode > ARRAY_CARDINALITY(errcodes))
+        errcode = AUG_EINTERNAL;
+    return errcodes[errcode];
+}
+
+const char *aug_error_minor_message(struct augeas *aug) {
+    return aug->error->minor_details;
+}
+
+const char *aug_error_details(struct augeas *aug) {
+    return aug->error->details;
+}
 
 /*
  * Local variables:
