@@ -29,6 +29,7 @@
 #include "syntax.h"
 #include "memory.h"
 #include "transform.h"
+#include "errcode.h"
 
 #define UNIMPL_BODY(name)                       \
     {                                           \
@@ -84,6 +85,9 @@ static struct value *make_exn_lns_error(struct info *info,
                                         const char *text) {
     struct value *v;
 
+    if (HAS_ERR(info))
+        return exn_error();
+
     v = make_exn_value(ref(info), "%s", err->message);
     if (err->lens != NULL) {
         char *s = format_info(err->lens->info);
@@ -92,9 +96,11 @@ static struct value *make_exn_lns_error(struct info *info,
     }
     if (err->pos >= 0) {
         char *pos = format_pos(text, err->pos);
+        size_t line, ofs;
+        calc_line_ofs(text, err->pos, &line, &ofs);
         exn_printf_line(v,
-                        "Error encountered here (%d characters into string)",
-                        err->pos);
+                     "Error encountered at %d:%d (%d characters into string)",
+                        (int) line, (int) ofs, err->pos);
         if (pos != NULL)
             exn_printf_line(v, "%s", pos);
         free(pos);
@@ -145,7 +151,7 @@ static struct value *lens_get(struct info *info, struct value *l,
     const char *text = str->string->str;
 
     struct tree *tree = lns_get(info, l->lens, text, &err);
-    if (err == NULL) {
+    if (err == NULL && ! HAS_ERR(info)) {
         v = make_value(V_TREE, ref(info));
         v->origin = make_tree_origin(tree);
     } else {
@@ -182,7 +188,7 @@ static struct value *lens_put(struct info *info, struct value *l,
             str->string->str, &err);
     close_memstream(&ms);
 
-    if (err == NULL) {
+    if (err == NULL && ! HAS_ERR(info)) {
         v = make_value(V_STRING, ref(info));
         v->string = make_string(ms.buf);
     } else {
@@ -212,7 +218,7 @@ static struct value *tree_set_glue(struct info *info, struct value *path,
         fake = tree->origin->children;
     }
 
-    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+    if (pathx_parse(tree->origin, NULL, path->string->str, true, NULL, &p)
         != PATHX_NOERROR) {
         result = make_pathx_exn(ref(info), p);
         goto done;
@@ -235,6 +241,47 @@ static struct value *tree_set_glue(struct info *info, struct value *path,
     return result;
 }
 
+/* V_STRING -> V_TREE -> V_TREE */
+static struct value *tree_clear_glue(struct info *info, struct value *path,
+                                     struct value *tree) {
+    // FIXME: This only works if TREE is not referenced more than once;
+    // otherwise we'll have some pretty weird semantics, and would really
+    // need to copy TREE first
+    assert(path->tag == V_STRING);
+    assert(tree->tag == V_TREE);
+
+    struct tree *fake = NULL;
+    struct pathx *p = NULL;
+    struct value *result = NULL;
+
+    if (tree->origin->children == NULL) {
+        tree->origin->children = make_tree(NULL, NULL, tree->origin, NULL);
+        fake = tree->origin->children;
+    }
+
+    if (pathx_parse(tree->origin, NULL, path->string->str, true, NULL, &p)
+        != PATHX_NOERROR) {
+        result = make_pathx_exn(ref(info), p);
+        goto done;
+    }
+
+    if (tree_set(p, NULL) == NULL) {
+        result = make_exn_value(ref(info),
+                                "Tree set of %s to NULL failed",
+                                path->string->str);
+        goto done;
+    }
+    if (fake != NULL) {
+        list_remove(fake, tree->origin->children);
+        free_tree(fake);
+    }
+    result = ref(tree);
+
+ done:
+    free_pathx(p);
+    return result;
+}
+
 static struct value *tree_insert_glue(struct info *info, struct value *label,
                                       struct value *path, struct value *tree,
                                       int before) {
@@ -249,7 +296,7 @@ static struct value *tree_insert_glue(struct info *info, struct value *label,
     struct pathx *p = NULL;
     struct value *result = NULL;
 
-    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+    if (pathx_parse(tree->origin, NULL, path->string->str, true, NULL, &p)
         != PATHX_NOERROR) {
         result = make_pathx_exn(ref(info), p);
         goto done;
@@ -296,7 +343,7 @@ static struct value *tree_rm_glue(struct info *info,
     struct pathx *p = NULL;
     struct value *result = NULL;
 
-    if (pathx_parse(tree->origin, path->string->str, true, NULL, &p)
+    if (pathx_parse(tree->origin, NULL, path->string->str, true, NULL, &p)
         != PATHX_NOERROR) {
         result = make_pathx_exn(ref(info), p);
         goto done;
@@ -371,7 +418,7 @@ static struct value *sys_read_file(struct info *info, struct value *n) {
     assert(n->tag == V_STRING);
     char *str = NULL;
 
-    str = read_file(n->string->str);
+    str = xread_file(n->string->str);
     if (str == NULL) {
         char error_buf[1024];
         const char *errmsg;
@@ -386,39 +433,64 @@ static struct value *sys_read_file(struct info *info, struct value *n) {
     return v;
 }
 
-struct module *builtin_init(void) {
+/* V_LENS -> V_LENS */
+static struct value *lns_check_rec_glue(struct info *info,
+                                        struct value *l, struct value *r) {
+    assert(l->tag == V_LENS);
+    assert(r->tag == V_LENS);
+    int check = info->error->aug->flags & AUG_TYPE_CHECK;
+
+    return lns_check_rec(info, l->lens, r->lens, check);
+}
+
+struct module *builtin_init(struct error *error) {
     struct module *modl = module_create("Builtin");
-    define_native(modl, "gensym", 1, gensym, T_STRING, T_STRING);
+    int r;
+
+#define DEFINE_NATIVE(modl, name, nargs, impl, types ...)               \
+    r = define_native(error, modl, name, nargs, impl, ##types);         \
+    if (r < 0) goto error;
+
+    DEFINE_NATIVE(modl, "gensym", 1, gensym, T_STRING, T_STRING);
+
     /* Primitive lenses */
-    define_native(modl, "del",     2, lns_del, T_REGEXP, T_STRING, T_LENS);
-    define_native(modl, "store",   1, lns_store, T_REGEXP, T_LENS);
-    define_native(modl, "key",     1, lns_key, T_REGEXP, T_LENS);
-    define_native(modl, "label",   1, lns_label, T_STRING, T_LENS);
-    define_native(modl, "seq",     1, lns_seq, T_STRING, T_LENS);
-    define_native(modl, "counter", 1, lns_counter, T_STRING, T_LENS);
+    DEFINE_NATIVE(modl, "del",     2, lns_del, T_REGEXP, T_STRING, T_LENS);
+    DEFINE_NATIVE(modl, "store",   1, lns_store, T_REGEXP, T_LENS);
+    DEFINE_NATIVE(modl, "key",     1, lns_key, T_REGEXP, T_LENS);
+    DEFINE_NATIVE(modl, "label",   1, lns_label, T_STRING, T_LENS);
+    DEFINE_NATIVE(modl, "seq",     1, lns_seq, T_STRING, T_LENS);
+    DEFINE_NATIVE(modl, "counter", 1, lns_counter, T_STRING, T_LENS);
     /* Applying lenses (mostly for tests) */
-    define_native(modl, "get",     2, lens_get, T_LENS, T_STRING, T_TREE);
-    define_native(modl, "put",     3, lens_put, T_LENS, T_TREE, T_STRING,
-                                               T_STRING);
+    DEFINE_NATIVE(modl, "get",     2, lens_get, T_LENS, T_STRING, T_TREE);
+    DEFINE_NATIVE(modl, "put",     3, lens_put, T_LENS, T_TREE, T_STRING,
+                  T_STRING);
     /* Tree manipulation used by the PUT tests */
-    define_native(modl, "set", 3, tree_set_glue, T_STRING, T_STRING, T_TREE,
+    DEFINE_NATIVE(modl, "set", 3, tree_set_glue, T_STRING, T_STRING, T_TREE,
                                                  T_TREE);
-    define_native(modl, "rm", 2, tree_rm_glue, T_STRING, T_TREE, T_TREE);
-    define_native(modl, "insa", 3, tree_insa_glue, T_STRING, T_STRING, T_TREE,
+    DEFINE_NATIVE(modl, "clear", 2, tree_clear_glue, T_STRING, T_TREE,
+                                                 T_TREE);
+    DEFINE_NATIVE(modl, "rm", 2, tree_rm_glue, T_STRING, T_TREE, T_TREE);
+    DEFINE_NATIVE(modl, "insa", 3, tree_insa_glue, T_STRING, T_STRING, T_TREE,
                                                    T_TREE);
-    define_native(modl, "insb", 3, tree_insb_glue, T_STRING, T_STRING, T_TREE,
+    DEFINE_NATIVE(modl, "insb", 3, tree_insb_glue, T_STRING, T_STRING, T_TREE,
                                                    T_TREE);
     /* Transforms and filters */
-    define_native(modl, "incl", 1, xform_incl, T_STRING, T_FILTER);
-    define_native(modl, "excl", 1, xform_excl, T_STRING, T_FILTER);
-    define_native(modl, "transform", 2, xform_transform, T_LENS, T_FILTER,
+    DEFINE_NATIVE(modl, "incl", 1, xform_incl, T_STRING, T_FILTER);
+    DEFINE_NATIVE(modl, "excl", 1, xform_excl, T_STRING, T_FILTER);
+    DEFINE_NATIVE(modl, "transform", 2, xform_transform, T_LENS, T_FILTER,
                                                          T_TRANSFORM);
+    DEFINE_NATIVE(modl, LNS_CHECK_REC_NAME,
+                  2, lns_check_rec_glue, T_LENS, T_LENS, T_LENS);
+
     /* System functions */
     struct module *sys = module_create("Sys");
     modl->next = sys;
-    define_native(sys, "getenv", 1, sys_getenv, T_STRING, T_STRING);
-    define_native(sys, "read_file", 1, sys_read_file, T_STRING, T_STRING);
+    DEFINE_NATIVE(sys, "getenv", 1, sys_getenv, T_STRING, T_STRING);
+    DEFINE_NATIVE(sys, "read_file", 1, sys_read_file, T_STRING, T_STRING);
     return modl;
+ error:
+    unref(modl, module);
+    return NULL;
 }
 
 /*
