@@ -25,6 +25,7 @@
 
 #include "list.h"
 #include "datadir.h"
+#include "augeas.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -34,6 +35,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <assert.h>
+#include <locale.h>
 
 /*
  * Various parameters about env vars, special tree nodes etc.
@@ -212,6 +214,10 @@ int pathjoin(char **path, int nseg, ...);
 
 #define MEMZERO(ptr, n) memset((ptr), 0, (n) * sizeof(*(ptr)));
 
+#define MEMCPY(dest, src, n) memcpy((dest), (src), (n) * sizeof(*(src)))
+
+#define MEMMOVE(dest, src, n) memmove((dest), (src), (n) * sizeof(*(src)))
+
 /**
  * TODO:
  *
@@ -252,17 +258,33 @@ int print_chars(FILE *out, const char *text, int cnt);
 void print_pos(FILE *out, const char *text, int pos);
 char *format_pos(const char *text, int pos);
 
-/* Function: read_file
+/* Function: xread_file
  * Read the contents of file PATH and return them as one long string. The
  * caller must free the result. Return NULL if any error occurs.
  */
-char* read_file(const char *path);
+char* xread_file(const char *path);
 
 /* Get the error message for ERRNUM in a threadsafe way. Based on libvirt's
  * virStrError
  */
 const char *xstrerror(int errnum, char *buf, size_t len);
 
+/* Like asprintf, but set *STRP to NULL on error */
+int xasprintf(char **strp, const char *format, ...);
+
+/* Calculate line and column number of character POS in TEXT */
+void calc_line_ofs(const char *text, size_t pos, size_t *line, size_t *ofs);
+
+/* Take the first LEN characters from the regexp *U and expand any
+ * character ranges in it. The expanded regexp, if expansion is necessary,
+ * is in U, and the old string is freed. If expansion is not needed or an
+ * error happens, U will be unchanged.
+ *
+ * Return 0 if expansion is not necessary, -1 if an error occurs, and 1 if
+ * expansion was needed.
+ */
+int regexp_c_locale(char **u, size_t *len);
+
 /* Struct: augeas
  * The data structure representing a connection to Augeas. */
 struct augeas {
@@ -275,8 +297,28 @@ struct augeas {
     char             *modpathz;   /* The search path for modules as a
                                      glibc argz vector */
     struct pathx_symtab *symtab;
+    struct error        *error;
+    uint                api_entries;  /* Number of entries through a public
+                                       * API, 0 when called from outside */
+#if HAVE_USELOCALE
+    /* On systems that have a uselocale call, we switch to the C locale
+     * on entry into API functions, and back to the old user locale
+     * on exit.
+     * FIXME: We need some solution for systems without uselocale, like
+     * setlocale + critical section, though that is very heavy-handed
+     */
+    locale_t            c_locale;
+    locale_t            user_locale;
+#endif
 };
 
+static inline struct error *err_of_aug(const struct augeas *aug) {
+    return ((struct augeas *) aug)->error;
+}
+
+/* Used by augparse for loading tests */
+int __aug_load_module_file(struct augeas *aug, const char *filename);
+
 /* Struct: tree
  * An entry in the global config tree. The data structure allows associating
  * values with interior nodes, but the API currently marks that as an error.
@@ -334,6 +376,29 @@ int dump_tree(FILE *out, struct tree *tree);
 int tree_equal(const struct tree *t1, const struct tree *t2);
 char *path_expand(struct tree *tree, const char *ppath);
 char *path_of_tree(struct tree *tree);
+/* Return first child with label LABEL or NULL */
+struct tree *tree_child(struct tree *tree, const char *label);
+/* Return first existing child with label LABEL or create one. Return NULL
+ * when allocation fails */
+struct tree *tree_child_cr(struct tree *tree, const char *label);
+/* Create a path in the tree; nodes along the path are looked up with
+ * tree_child_cr */
+struct tree *tree_path_cr(struct tree *tree, int n, ...);
+/* Set the value of TREE and update dirty flags */
+int tree_set_value(struct tree *tree, const char *value);
+/* Cleanly remove all children of TREE, but leave TREE itself unchanged */
+void tree_unlink_children(struct augeas *aug, struct tree *tree);
+/* Find the node matching PATH.
+ * Returns the node or NULL on error
+ * Errors: EMMATCH - more than one node matches PATH
+ *         ENOMEM  - allocation error
+ */
+struct tree *tree_find(struct augeas *aug, const char *path);
+/* Find the node matching PATH. Expand the tree to contain such a node if
+ * none exists.
+ * Returns the node or NULL on error
+ */
+struct tree *tree_find_cr(struct augeas *aug, const char *path);
 
 /* Struct: memstream
  * Wrappers to simulate OPEN_MEMSTREAM where that's not available. The
@@ -385,7 +450,9 @@ typedef enum {
     PATHX_ETYPE,
     PATHX_ENOVAR,
     PATHX_EEND,
-    PATHX_ENONODES
+    PATHX_ENONODES,
+    PATHX_EARITY,
+    PATHX_EREGEXP
 } pathx_errcode_t;
 
 struct pathx;
@@ -401,23 +468,50 @@ const char *pathx_error(struct pathx *pathx, const char **txt, int *pos);
  *
  * Returns 0 on success, and -1 on error
  */
-int pathx_parse(const struct tree *origin, const char *path,
+int pathx_parse(const struct tree *origin,
+                struct error *err,
+                const char *path,
                 bool need_nodeset,
                 struct pathx_symtab *symtab,
                 struct pathx **px);
+/* Return the error struct that was passed into pathx_parse */
+struct error *err_of_pathx(struct pathx *px);
 struct tree *pathx_first(struct pathx *path);
 struct tree *pathx_next(struct pathx *path);
+/* Return -1 if evalutating PATH runs into trouble, otherwise return the
+ * number of nodes matching PATH and set MATCH to the first matching
+ * node */
 int pathx_find_one(struct pathx *path, struct tree **match);
 int pathx_expand_tree(struct pathx *path, struct tree **tree);
 void free_pathx(struct pathx *path);
 
-int pathx_symtab_init(struct pathx_symtab **symtab);
+struct pathx_symtab *pathx_get_symtab(struct pathx *pathx);
 int pathx_symtab_define(struct pathx_symtab **symtab,
                         const char *name, struct pathx *px);
+/* Returns 1 on success, and -1 when out of memory */
+int pathx_symtab_assign_tree(struct pathx_symtab **symtab, const char *name,
+                             struct tree *tree);
 int pathx_symtab_undefine(struct pathx_symtab **symtab, const char *name);
-void pathx_symtab_remove_descendants(struct pathx *pathx,
+void pathx_symtab_remove_descendants(struct pathx_symtab *symtab,
                                      const struct tree *tree);
 void free_symtab(struct pathx_symtab *symtab);
+
+/* Debug helpers, all defined in internal.c. When ENABLE_DEBUG is not
+ * set, they compile to nothing.
+ */
+#  if ENABLE_DEBUG
+  /* Return true if debugging for CATEGORY is turned on */
+  bool debugging(const char *category);
+  /* Format the arguments into a file name, prepend it with the directory
+   * from the environment variable AUGEAS_DEBUG_DIR, and open the file for
+   * writing.
+  */
+  FILE *debug_fopen(const char *format, ...)
+    ATTRIBUTE_FORMAT(printf, 1, 2);
+#  else
+#    define debugging(facility) (0)
+#    define debug_fopen(format ...) (NULL)
+#  endif
 #endif
 
 
