@@ -27,6 +27,10 @@
 #include <memory.h>
 #include <ctype.h>
 
+#include "ref.h"
+#include "regexp.h"
+#include "errcode.h"
+
 static const char *const errcodes[] = {
     "no error",
     "empty name",
@@ -42,7 +46,9 @@ static const char *const errcodes[] = {
     "type error",       /* PATHX_ETYPE */
     "undefined variable",               /* PATHX_ENOVAR */
     "garbage at end of path expression",/* PATHX_EEND */
-    "can not expand tree from empty nodeset"  /* PATHX_ENONODES */
+    "can not expand tree from empty nodeset",  /* PATHX_ENONODES */
+    "wrong number of arguments in function call", /* PATHX_EARITY */
+    "invalid regular expression"        /* PATHX_EREGEXP */
 };
 
 /*
@@ -54,7 +60,8 @@ enum type {
     T_NODESET,
     T_BOOLEAN,
     T_NUMBER,
-    T_STRING
+    T_STRING,
+    T_REGEXP
 };
 
 enum expr_tag {
@@ -76,7 +83,9 @@ enum binary_op {
     OP_MINUS,      /* '-'  */
     OP_STAR,       /* '*'  */
     OP_AND,        /* 'and' */
-    OP_OR          /* 'or' */
+    OP_OR,         /* 'or' */
+    OP_RE_MATCH,   /* '=~' */
+    OP_UNION       /* '|' */
 };
 
 struct pred {
@@ -91,7 +100,9 @@ enum axis {
     DESCENDANT_OR_SELF,
     PARENT,
     ANCESTOR,
-    ROOT
+    ROOT,
+    PRECEDING_SIBLING,
+    FOLLOWING_SIBLING
 };
 
 /* This array is indexed by enum axis */
@@ -102,7 +113,9 @@ static const char *const axis_names[] = {
     "descendant-or-self",
     "parent",
     "ancestor",
-    "root"
+    "root",
+    "preceding-sibling",
+    "following-sibling"
 };
 
 static const char *const axis_sep = "::";
@@ -159,6 +172,7 @@ struct value {
         int              number;      /* T_NUMBER  */
         char            *string;      /* T_STRING  */
         bool             boolval;     /* T_BOOLEAN */
+        struct regexp   *regexp;      /* T_REGEXP  */
     };
 };
 
@@ -229,6 +243,9 @@ struct state {
     struct locpath_trace *locpath_trace;
     /* Symbol table for variable lookups */
     struct pathx_symtab *symtab;
+    /* Error structure, used to communicate errors to struct augeas;
+     * we never own this structure, and therefore never free it */
+    struct error        *error;
 };
 
 /* We consider NULL and the empty string to be equal */
@@ -257,8 +274,10 @@ static void func_last(struct state *state);
 static void func_position(struct state *state);
 static void func_count(struct state *state);
 static void func_label(struct state *state);
+static void func_regexp(struct state *state);
 
 static const enum type const count_arg_types[] = { T_NODESET };
+static const enum type const regexp_arg_types[] = { T_STRING };
 
 static const struct func builtin_funcs[] = {
     { .name = "last", .arity = 0, .type = T_NUMBER, .arg_types = NULL,
@@ -269,7 +288,10 @@ static const struct func builtin_funcs[] = {
       .impl = func_label },
     { .name = "count", .arity = 1, .type = T_NUMBER,
       .arg_types = count_arg_types,
-      .impl = func_count }
+      .impl = func_count },
+    { .name = "regexp", .arity = 1, .type = T_REGEXP,
+      .arg_types = regexp_arg_types,
+      .impl =func_regexp }
 };
 
 #define CHECK_ERROR                                                     \
@@ -387,6 +409,9 @@ static void release_value(struct value *v) {
     case T_BOOLEAN:
     case T_NUMBER:
         break;
+    case T_REGEXP:
+        unref(v->regexp, regexp);
+        break;
     default:
         assert(0);
     }
@@ -426,7 +451,9 @@ static struct nodeset *make_nodeset(struct state *state) {
 
 static void ns_add(struct nodeset *ns, struct tree *node,
                    struct state *state) {
-    // FIXME: Need to uniquify NS
+    for (int i=0; i < ns->used; i++)
+        if (ns->nodes[i] == node)
+            return;
     if (ns->used >= ns->size) {
         size_t size = 2 * ns->size;
         if (size < 10) size = 10;
@@ -452,7 +479,7 @@ clone_nodeset(struct nodeset *ns, struct state *state)
         return NULL;
     }
     clone->used = ns->used;
-    clone->size = ns->size;
+    clone->size = ns->used;
     for (int i=0; i < ns->used; i++)
         clone->nodes[i] = ns->nodes[i];
     return clone;
@@ -503,6 +530,9 @@ static value_ind_t clone_value(struct value *v, struct state *state) {
     case T_BOOLEAN:
         clone->boolval = v->boolval;
         break;
+    case T_REGEXP:
+        clone->regexp = ref(v->regexp);
+        break;
     default:
         assert(0);
     }
@@ -597,6 +627,35 @@ static void func_label(struct state *state) {
     push_value(vind, state);
 }
 
+static void func_regexp(struct state *state) {
+    value_ind_t vind = make_value(T_REGEXP, state);
+
+    CHECK_ERROR;
+
+    struct value *str = pop_value(state);
+    assert(str->tag == T_STRING);
+
+    char *pat = strdup(str->string);
+    if (pat == NULL) {
+        STATE_ENOMEM;
+        return;
+    }
+
+    struct regexp *rx = make_regexp(NULL, pat, 0);
+    if (rx == NULL) {
+        FREE(pat);
+        STATE_ENOMEM;
+        return;
+    }
+    if (regexp_compile(rx) < 0) {
+        STATE_ERROR(state, PATHX_EREGEXP);
+        unref(rx, regexp);
+        return;
+    }
+    state->value_pool[vind].regexp = rx;
+    push_value(vind, state);
+}
+
 static bool coerce_to_bool(struct value *v) {
     switch (v->tag) {
     case T_NODESET:
@@ -611,6 +670,8 @@ static bool coerce_to_bool(struct value *v) {
     case T_STRING:
         return strlen(v->string) > 0;
         break;
+    case T_REGEXP:
+        return true;
     default:
         assert(0);
     }
@@ -745,6 +806,64 @@ static void eval_and_or(struct state *state, enum binary_op op) {
         push_boolean_value(left || right, state);
 }
 
+static bool eval_re_match_str(struct state *state, struct regexp *rx,
+                              const char *str) {
+    int r;
+
+    if (str == NULL)
+        str = "";
+
+    r = regexp_match(rx, str, strlen(str), 0, NULL);
+    if (r == -2) {
+        STATE_ERROR(state, PATHX_EINTERNAL);
+    } else if (r == -3) {
+        /* We should never get this far; func_regexp should catch
+         * invalid regexps */
+        assert(false);
+    }
+    return r == strlen(str);
+}
+
+static void eval_union(struct state *state) {
+    value_ind_t vind = make_value(T_NODESET, state);
+    struct value *r = pop_value(state);
+    struct value *l = pop_value(state);
+    struct nodeset *res = NULL;
+
+    assert(l->tag == T_NODESET);
+    assert(r->tag == T_NODESET);
+
+    CHECK_ERROR;
+
+    res = clone_nodeset(l->nodeset, state);
+    CHECK_ERROR;
+    for (int i=0; i < r->nodeset->used; i++) {
+        ns_add(res, r->nodeset->nodes[i], state);
+        CHECK_ERROR;
+    }
+    state->value_pool[vind].nodeset = res;
+    push_value(vind, state);
+}
+
+static void eval_re_match(struct state *state) {
+    struct value *rx  = pop_value(state);
+    struct value *v = pop_value(state);
+
+    bool result = false;
+
+    if (v->tag == T_STRING) {
+        result = eval_re_match_str(state, rx->regexp, v->string);
+        CHECK_ERROR;
+    } else if (v->tag == T_NODESET) {
+        for (int i=0; i < v->nodeset->used && result == false; i++) {
+            struct tree *t = v->nodeset->nodes[i];
+            result = eval_re_match_str(state, rx->regexp, t->value);
+            CHECK_ERROR;
+        }
+    }
+    push_boolean_value(result, state);
+}
+
 static void eval_binary(struct expr *expr, struct state *state) {
     eval_expr(expr->left, state);
     eval_expr(expr->right, state);
@@ -778,6 +897,12 @@ static void eval_binary(struct expr *expr, struct state *state) {
     case OP_OR:
         eval_and_or(state, expr->op);
         break;
+    case OP_UNION:
+        eval_union(state);
+        break;
+    case OP_RE_MATCH:
+        eval_re_match(state);
+        break;
     default:
         assert(0);
     }
@@ -793,8 +918,10 @@ static void eval_app(struct expr *expr, struct state *state) {
     expr->func->impl(state);
 }
 
-static int eval_pred(struct expr *expr, struct state *state) {
+static bool eval_pred(struct expr *expr, struct state *state) {
     eval_expr(expr, state);
+    CHECK_ERROR_RET0;
+
     struct value *v = pop_value(state);
     switch(v->tag) {
     case T_BOOLEAN:
@@ -831,7 +958,9 @@ static void ns_filter(struct nodeset *ns, struct pred *predicates,
         state->ctx_pos = 1;
         for (int i=0; i < ns->used; state->ctx_pos++) {
             state->ctx = ns->nodes[i];
-            if (eval_pred(predicates->exprs[p], state)) {
+            bool match = eval_pred(predicates->exprs[p], state);
+            CHECK_ERROR;
+            if (match) {
                 i+=1;
             } else {
                 ns_remove(ns, i);
@@ -887,6 +1016,7 @@ static void ns_from_locpath(struct locpath *lp, uint *maxns,
                 ns_add(next, node, state);
         }
         ns_filter(next, step->predicates, state);
+        CHECK_ERROR;
         cur_ns += 1;
     }
 
@@ -1055,6 +1185,10 @@ static void check_app(struct expr *expr, struct state *state) {
  * '+', '-', '*': T_NUMBER -> T_NUMBER -> T_NUMBER
  *
  * 'and', 'or': T_BOOLEAN -> T_BOOLEAN -> T_BOOLEAN
+ * '=~'       : T_STRING  -> T_REGEXP  -> T_BOOLEAN
+ *              T_NODESET -> T_REGEXP  -> T_BOOLEAN
+ *
+ * '|'        : T_NODESET -> T_NODESET -> T_NODESET
  *
  * Any type can be coerced to T_BOOLEAN (see coerce_to_bool)
  */
@@ -1090,11 +1224,19 @@ static void check_binary(struct expr *expr, struct state *state) {
         ok =  (l == T_NUMBER && r == T_NUMBER);
         res = T_NUMBER;
         break;
+    case OP_UNION:
+        ok = (l == T_NODESET && r == T_NODESET);
+        res = T_NODESET;
+        break;
     case OP_AND:
     case OP_OR:
         ok = 1;
         res = T_BOOLEAN;
         break;
+    case OP_RE_MATCH:
+        ok = ((l == T_STRING || l == T_NODESET) && r == T_REGEXP);
+        res = T_BOOLEAN;
+        break;
     default:
         assert(0);
     }
@@ -1585,7 +1727,7 @@ static void parse_function_call(struct state *state) {
     }
 
     if (nargs != func->arity) {
-        STATE_ERROR(state, PATHX_EDELIM);
+        STATE_ERROR(state, PATHX_EARITY);
         return;
     }
 
@@ -1757,14 +1899,27 @@ static void parse_path_expr(struct state *state) {
 }
 
 /*
- * MultiplicativeExpr ::= PathExpr ('*' PathExpr)*
+ * UnionExpr ::= PathExpr ('|' PathExpr)*
  */
-static void parse_multiplicative_expr(struct state *state) {
+static void parse_union_expr(struct state *state) {
     parse_path_expr(state);
     CHECK_ERROR;
-    while (match(state, '*')) {
+    while (match(state, '|')) {
         parse_path_expr(state);
         CHECK_ERROR;
+        push_new_binary_op(OP_UNION, state);
+    }
+}
+
+/*
+ * MultiplicativeExpr ::= UnionExpr ('*' UnionExpr)*
+ */
+static void parse_multiplicative_expr(struct state *state) {
+    parse_union_expr(state);
+    CHECK_ERROR;
+    while (match(state, '*')) {
+        parse_union_expr(state);
+        CHECK_ERROR;
         push_new_binary_op(OP_STAR, state);
     }
 }
@@ -1808,13 +1963,20 @@ static void parse_relational_expr(struct state *state) {
 }
 
 /*
- * EqualityExpr ::= RelationalExpr (EqualityOp RelationalExpr)?
+ * EqualityExpr ::= RelationalExpr (EqualityOp RelationalExpr)? | ReMatchExpr
  * EqualityOp ::= "=" | "!="
+ * ReMatchExpr ::= RelationalExpr "=~" RelationalExpr
  */
 static void parse_equality_expr(struct state *state) {
     parse_relational_expr(state);
     CHECK_ERROR;
-    if (*state->pos == '=' ||
+    if (*state->pos == '=' && state->pos[1] == '~') {
+        state->pos += 2;
+        skipws(state);
+        parse_relational_expr(state);
+        CHECK_ERROR;
+        push_new_binary_op(OP_RE_MATCH, state);
+    } else if (*state->pos == '=' ||
         (*state->pos == '!' && state->pos[1] == '=')) {
         enum binary_op op = (*state->pos == '=') ? OP_EQ : OP_NEQ;
         state->pos += (op == OP_EQ) ? 1 : 2;
@@ -1864,7 +2026,38 @@ static void parse_expr(struct state *state) {
     parse_or_expr(state);
 }
 
-int pathx_parse(const struct tree *tree, const char *txt,
+static void store_error(struct pathx *pathx) {
+    const char *pathx_msg = NULL;
+    const char *path = pathx->state->txt;
+    const pathx_errcode_t errcode = pathx->state->errcode;
+    struct error *err = pathx->state->error;
+
+    char *pos_str = NULL;
+    int pos;
+
+    if (err == NULL || errcode == PATHX_NOERROR || err->code != AUG_NOERROR)
+        return;
+
+    pathx_msg = pathx_error(pathx, NULL, &pos);
+
+    if (ALLOC_N(pos_str, strlen(path) + 4) >= 0) {
+        strncpy(pos_str, path, pos);
+        strcat(pos_str, "|=|");
+        strcat(pos_str, path + pos);
+    }
+
+    err->code = errcode == PATHX_ENOMEM ? AUG_ENOMEM : AUG_EPATHX;
+    err->minor = errcode;
+    err->details = pos_str;
+    pos_str = NULL;
+    err->minor_details = pathx_msg;
+
+    FREE(pos_str);
+}
+
+int pathx_parse(const struct tree *tree,
+                struct error *err,
+                const char *txt,
                 bool need_nodeset,
                 struct pathx_symtab *symtab,
                 struct pathx **pathx) {
@@ -1873,22 +2066,20 @@ int pathx_parse(const struct tree *tree, const char *txt,
     *pathx = NULL;
 
     if (ALLOC(*pathx) < 0)
-        return PATHX_ENOMEM;
+        goto oom;
 
     (*pathx)->origin = (struct tree *) tree;
 
     /* Set up state */
-    if (ALLOC((*pathx)->state) < 0) {
-        free_pathx(*pathx);
-        *pathx = NULL;
-        return PATHX_ENOMEM;
-    }
+    if (ALLOC((*pathx)->state) < 0)
+        goto oom;
     state = (*pathx)->state;
 
     state->errcode = PATHX_NOERROR;
     state->txt = txt;
     state->pos = txt;
     state->symtab = symtab;
+    state->error = err;
 
     if (ALLOC_N(state->value_pool, 8) < 0) {
         STATE_ENOMEM;
@@ -1926,7 +2117,14 @@ int pathx_parse(const struct tree *tree, const char *txt,
     }
 
  done:
+    store_error(*pathx);
     return state->errcode;
+ oom:
+    free_pathx(*pathx);
+    *pathx = NULL;
+    if (err != NULL)
+        err->code = AUG_ENOMEM;
+    return PATHX_ENOMEM;
 }
 
 /*************************************************************************
@@ -1937,6 +2135,16 @@ static bool step_matches(struct step *step, struct tree *tree) {
     return (step->name == NULL || streqx(step->name, tree->label));
 }
 
+static struct tree *tree_prev(struct tree *pos) {
+    struct tree *node = NULL;
+    if (pos != pos->parent->children) {
+        for (node = pos->parent->children;
+             node->next != pos;
+             node = node->next);
+    }
+    return node;
+}
+
 static struct tree *step_first(struct step *step, struct tree *ctx) {
     struct tree *node = NULL;
     switch (step->axis) {
@@ -1957,6 +2165,12 @@ static struct tree *step_first(struct step *step, struct tree *ctx) {
             ctx = ctx->parent;
         node = ctx;
         break;
+    case PRECEDING_SIBLING:
+        node = tree_prev(ctx);
+        break;
+    case FOLLOWING_SIBLING:
+        node = ctx->next;
+        break;
     default:
         assert(0);
     }
@@ -2000,6 +2214,12 @@ static struct tree *step_next(struct step *step, struct tree *ctx,
             else
                 node = node->parent;
             break;
+        case PRECEDING_SIBLING:
+            node = tree_prev(node);
+            break;
+        case FOLLOWING_SIBLING:
+            node = node->next;
+            break;
         default:
             assert(0);
         }
@@ -2037,7 +2257,7 @@ struct tree *pathx_first(struct pathx *pathx) {
         struct value *v = pathx_eval(pathx);
 
         if (HAS_ERROR(pathx->state))
-            return NULL;
+            goto error;
         assert(v->tag == T_NODESET);
         pathx->nodeset = v->nodeset;
     }
@@ -2046,6 +2266,9 @@ struct tree *pathx_first(struct pathx *pathx) {
         return NULL;
     else
         return pathx->nodeset->nodes[0];
+ error:
+    store_error(pathx);
+    return NULL;
 }
 
 /* Find a node in the tree that matches the longest prefix of PATH.
@@ -2152,19 +2375,19 @@ int pathx_expand_tree(struct pathx *path, struct tree **tree) {
         free_tree(first_child);
     }
     *tree = NULL;
+    store_error(path);
     return -1;
 }
 
 int pathx_find_one(struct pathx *path, struct tree **tree) {
     *tree = pathx_first(path);
-    if (*tree == NULL)
-        return 0;
-
-    if (pathx_next(path) != NULL) {
-        *tree = NULL;
+    if (HAS_ERROR(path->state))
         return -1;
-    }
-    return 1;
+    return path->nodeset->used;
+}
+
+struct error *err_of_pathx(struct pathx *px) {
+    return px->state->error;
 }
 
 const char *pathx_error(struct pathx *path, const char **txt, int *pos) {
@@ -2227,56 +2450,61 @@ void free_symtab(struct pathx_symtab *symtab) {
     }
 }
 
-int pathx_symtab_init(struct pathx_symtab **symtab) {
-    struct value *v = NULL;
-
-    *symtab = NULL;
-
-    if (ALLOC(v) < 0)
-        goto error;
-    v->tag = T_BOOLEAN;
-
-    *symtab = make_symtab(NULL, " unused ", v);
-    if (*symtab == NULL)
-        goto error;
-    return 0;
- error:
-    release_value(v);
-    free(v);
-    free_symtab(*symtab);
-    *symtab = NULL;
-    return -1;
+struct pathx_symtab *pathx_get_symtab(struct pathx *pathx) {
+    return pathx->state->symtab;
 }
 
-int pathx_symtab_define(struct pathx_symtab **symtab,
-                        const char *name, struct pathx *px) {
-    struct pathx_symtab *new;
-    struct value *value = NULL, *v = NULL;
-
-    value = pathx_eval(px);
-    if (HAS_ERROR(px->state))
-        goto error;
-
-    if (ALLOC(v) < 0)
-        goto error;
-    *v = *value;
-    value->tag = T_BOOLEAN;
+static int pathx_symtab_set(struct pathx_symtab **symtab,
+                            const char *name, struct value *v) {
+    int found = 0;
 
     list_for_each(tab, *symtab) {
         if (STREQ(tab->name, name)) {
             release_value(tab->value);
             free(tab->value);
             tab->value = v;
-            goto done;
+            found = 1;
+            break;
         }
     }
 
-    new = make_symtab(*symtab, name, v);
-    if (new == NULL)
+    if (!found) {
+        struct pathx_symtab *new = NULL;
+
+        new = make_symtab(*symtab, name, v);
+        if (new == NULL)
+            goto error;
+        *symtab = new;
+    }
+    return 0;
+ error:
+    return -1;
+}
+
+int pathx_symtab_define(struct pathx_symtab **symtab,
+                        const char *name, struct pathx *px) {
+    int r;
+    struct value *value = NULL, *v = NULL;
+    struct state *state = px->state;
+
+    value = pathx_eval(px);
+    if (HAS_ERROR(px->state))
         goto error;
 
-    *symtab = new;
- done:
+    if (ALLOC(v) < 0) {
+        STATE_ENOMEM;
+        goto error;
+    }
+
+    *v = *value;
+    value->tag = T_BOOLEAN;
+
+    r = pathx_symtab_set(symtab, name, v);
+    if (r < 0) {
+        STATE_ENOMEM;
+        goto error;
+    }
+
     if (v->tag == T_NODESET)
         return v->nodeset->used;
     else
@@ -2286,6 +2514,7 @@ int pathx_symtab_define(struct pathx_symtab **symtab,
     free(value);
     release_value(v);
     free(v);
+    store_error(px);
     return -1;
 }
 
@@ -2302,9 +2531,35 @@ int pathx_symtab_undefine(struct pathx_symtab **symtab, const char *name) {
     return 0;
 }
 
-void pathx_symtab_remove_descendants(struct pathx *pathx,
+int pathx_symtab_assign_tree(struct pathx_symtab **symtab,
+                             const char *name, struct tree *tree) {
+    struct value *v = NULL;
+    int r;
+
+    if (ALLOC(v) < 0)
+        goto error;
+
+    v->tag = T_NODESET;
+    if (ALLOC(v->nodeset) < 0)
+        goto error;
+    if (ALLOC_N(v->nodeset->nodes, 1) < 0)
+        goto error;
+    v->nodeset->used = 1;
+    v->nodeset->size = 1;
+    v->nodeset->nodes[0] = tree;
+
+    r = pathx_symtab_set(symtab, name, v);
+    if (r < 0)
+        goto error;
+    return 1;
+ error:
+    release_value(v);
+    free(v);
+    return -1;
+}
+
+void pathx_symtab_remove_descendants(struct pathx_symtab *symtab,
                                      const struct tree *tree) {
-    struct pathx_symtab *symtab = pathx->state->symtab;
     list_for_each(tab, symtab) {
         if (tab->value->tag != T_NODESET)
             continue;
