@@ -36,6 +36,7 @@
 #include "syntax.h"
 #include "augeas.h"
 #include "transform.h"
+#include "errcode.h"
 
 /* Extension of source files */
 #define AUG_EXT ".aug"
@@ -75,67 +76,60 @@ struct ctx {
     struct binding *local;
 };
 
-char *format_info(struct info *info) {
-    const char *fname;
-    char *result;
-    int r = 0;
-    int fl = info->first_line, ll = info->last_line;
-    int fc = info->first_column, lc = info->last_column;
-    fname = (info->filename != NULL) ? info->filename->str : "(unknown file)";
+static void format_error(struct info *info, aug_errcode_t code,
+                         const char *format, va_list ap) {
+    struct error *error = info->error;
+    char *si = NULL, *sf = NULL, *sd = NULL;
+    int r;
 
-    if (fl > 0) {
-        if (fl == ll) {
-            if (fc == lc) {
-                r = asprintf(&result, "%s:%d.%d", fname, fl, fc);
-            } else {
-                r = asprintf(&result, "%s:%d.%d-.%d", fname, fl, fc, lc);
-            }
-        } else {
-            r = asprintf(&result, "%s:%d.%d-%d.%d", fname, fl, fc, ll, lc);
-        }
-    }
-    return (r == -1) ? NULL : result;
-}
+    error->code = code;
+    /* Only syntax errors are cumulative */
+    if (code != AUG_ESYNTAX)
+        FREE(error->details);
 
-void print_info(FILE *out, struct info *info) {
-    fprintf(out, "%s:",
-            info->filename != NULL ? info->filename->str : "(unknown file)");
-    if (info->first_line > 0) {
-        if (info->first_line == info->last_line) {
-            if (info->first_column == info->last_column) {
-                fprintf(out, "%d.%d:", info->first_line, info->first_column);
-            } else {
-                fprintf(out, "%d.%d-.%d:", info->first_line,
-                        info->first_column, info->last_column);
-            }
-        } else {
-            fprintf(out, "%d.%d-%d.%d:",
-                    info->first_line, info->first_column,
-                    info->last_line, info->last_column);
-        }
+    si = format_info(info);
+    r = vasprintf(&sf, format, ap);
+    if (r < 0)
+        sf = NULL;
+    if (error->details != NULL) {
+        r = xasprintf(&sd, "%s\n%s%s", error->details,
+                      (si == NULL) ? "(no location)" : si,
+                      (sf == NULL) ? "(no details)" : sf);
+    } else {
+        r = xasprintf(&sd, "%s%s",
+                      (si == NULL) ? "(no location)" : si,
+                      (sf == NULL) ? "(no details)" : sf);
     }
+    if (r >= 0) {
+        free(error->details);
+        error->details = sd;
+    }
+    free(si);
+    free(sf);
 }
 
 void syntax_error(struct info *info, const char *format, ...) {
+    struct error *error = info->error;
     va_list ap;
 
-    print_info(stderr, info);
+    if (error->code != AUG_NOERROR && error->code != AUG_ESYNTAX)
+        return;
+
 	va_start(ap, format);
-    vfprintf(stderr, format, ap);
+    format_error(info, AUG_ESYNTAX, format, ap);
     va_end(ap);
-    fprintf(stderr, "\n");
 }
 
 void fatal_error(struct info *info, const char *format, ...) {
+    struct error *error = info->error;
     va_list ap;
 
-    fprintf(stderr, "Fatal internal error. Aborting. Details:\n");
-    print_info(stderr, info);
+    if (error->code == AUG_EINTERNAL)
+        return;
+
 	va_start(ap, format);
-    vfprintf(stderr, format, ap);
+    format_error(info, AUG_EINTERNAL, format, ap);
     va_end(ap);
-    fprintf(stderr, "\n");
-    abort();
 }
 
 void assert_error_at(const char *srcfile, int srclineno, struct info *info,
@@ -151,22 +145,6 @@ void assert_error_at(const char *srcfile, int srclineno, struct info *info,
     fprintf(stderr, "\n");
 }
 
-void free_string(struct string *string) {
-    if (string == NULL)
-        return;
-    assert(string->ref == 0);
-    free(string->str);
-    free(string);
-}
-
-void free_info(struct info *info) {
-    if (info == NULL)
-        return;
-    assert(info->ref == 0);
-    unref(info->filename, string);
-    free(info);
-}
-
 static void free_param(struct param *param) {
     if (param == NULL)
         return;
@@ -177,7 +155,7 @@ static void free_param(struct param *param) {
     free(param);
 }
 
-static void free_term(struct term *term) {
+void free_term(struct term *term) {
     if (term == NULL)
         return;
     assert(term->ref == 0);
@@ -302,7 +280,8 @@ void free_value(struct value *v) {
         unref(v->transform, transform);
         break;
     case V_NATIVE:
-        unref(v->native->type, type);
+        if (v->native)
+            unref(v->native->type, type);
         free(v->native);
         break;
     case V_CLOS:
@@ -327,61 +306,63 @@ void free_value(struct value *v) {
  */
 struct term *make_term(enum term_tag tag, struct info *info) {
   struct term *term;
-  make_ref(term);
-  term->tag = tag;
-  if (info != NULL)
+  if (make_ref(term) < 0) {
+      unref(info, info);
+  } else {
+      term->tag = tag;
       term->info = info;
+  }
   return term;
 }
 
 struct term *make_param(char *name, struct type *type, struct info *info) {
   struct term *term = make_term(A_FUNC, info);
-  make_ref(term->param);
+  if (term == NULL)
+      goto error;
+  make_ref_err(term->param);
   term->param->info = ref(term->info);
-  make_ref(term->param->name);
+  make_ref_err(term->param->name);
   term->param->name->str = name;
   term->param->type = type;
   return term;
+ error:
+  unref(term, term);
+  return NULL;
 }
 
 struct value *make_value(enum value_tag tag, struct info *info) {
-    struct value *value;
-    make_ref(value);
-    value->tag = tag;
-    value->info = info;
+    struct value *value = NULL;
+    if (make_ref(value) < 0) {
+        unref(info, info);
+    } else {
+        value->tag = tag;
+        value->info = info;
+    }
     return value;
 }
 
-struct string *make_string(char *str) {
-    struct string *string;
-    make_ref(string);
-    string->str = str;
-    return string;
-}
-
-struct string *dup_string(const char *str) {
-    struct string *string;
-    make_ref(string);
-    if (str == NULL)
-        string->str = strdup("");
-    else
-        string->str = strdup(str);
-    if (string->str == NULL)
-        unref(string, string);
-    return string;
-}
-
 struct term *make_app_term(struct term *lambda, struct term *arg,
                            struct info *info) {
   struct term *app = make_term(A_APP, info);
-  app->left = lambda;
-  app->right = arg;
+  if (app == NULL) {
+      unref(lambda, term);
+      unref(arg, term);
+  } else {
+      app->left = lambda;
+      app->right = arg;
+  }
   return app;
 }
 
 struct term *make_app_ident(char *id, struct term *arg, struct info *info) {
     struct term *ident = make_term(A_IDENT, ref(info));
     ident->ident = make_string(id);
+    if (ident->ident == NULL) {
+        unref(arg, term);
+        unref(info, info);
+        unref(ident, term);
+        return NULL;
+    }
     return make_app_term(ident, arg, info);
 }
 
@@ -397,12 +378,13 @@ struct term *build_func(struct term *params, struct term *exp) {
 
 /* Ownership is taken as needed */
 static struct value *make_closure(struct term *func, struct binding *bnds) {
-    struct value *v;
-    make_ref(v);
-    v->tag  = V_CLOS;
-    v->info = ref(func->info);
-    v->func = ref(func);
-    v->bindings = ref(bnds);
+    struct value *v = NULL;
+    if (make_ref(v) == 0) {
+        v->tag  = V_CLOS;
+        v->info = ref(func->info);
+        v->func = ref(func);
+        v->bindings = ref(bnds);
+    }
     return v;
 }
 
@@ -454,6 +436,18 @@ void exn_printf_line(struct value *exn, const char *format, ...) {
         exn_add_lines(exn, 1, line);
 }
 
+struct value *exn_error(void) {
+    static const struct exn exn = {
+        .info = NULL, .seen = 1, .error = 1,
+        .message = (char *) "Error during evaluation",
+        .nlines = 0, .lines = NULL };
+    static const struct value value = {
+        .ref = REF_MAX, /* Protect against being freed */
+        .info = NULL, .tag = V_EXN,
+        { .exn = (struct exn *) &exn } };
+    return (struct value *) &value;
+}
+
 /*
  * Modules
  */
@@ -652,6 +646,25 @@ static void dump_ctx(struct ctx *ctx) {
 /*
  * Values
  */
+static void print_tree(FILE *out, int indent, struct tree *tree) {
+    list_for_each(t, tree) {
+        for (int i=0; i < indent; i++) fputc(' ', out);
+        fprintf(out, "{ ");
+        if (t->label != NULL)
+            fprintf(out, "\"%s\"", t->label);
+        if (t->value != NULL)
+            fprintf(out, " = \"%s\"", t->value);
+        if (t->children != NULL) {
+            fputc('\n', out);
+            print_tree(out, indent + 2, t->children);
+            for (int i=0; i < indent; i++) fputc(' ', out);
+        } else {
+            fputc(' ', out);
+        }
+        fprintf(out, "}\n");
+    }
+}
+
 static void print_value(FILE *out, struct value *v) {
     if (v == NULL) {
         fprintf(out, "<null>");
@@ -667,11 +680,11 @@ static void print_value(FILE *out, struct value *v) {
         break;
     case V_LENS:
         fprintf(out, "<lens:");
-        print_info(stdout, v->lens->info);
+        print_info(out, v->lens->info);
         fprintf(out, ">");
         break;
     case V_TREE:
-        dump_tree(stdout, v->origin);
+        print_tree(out, 0, v->origin);
         break;
     case V_FILTER:
         fprintf(out, "<filter:");
@@ -682,24 +695,27 @@ static void print_value(FILE *out, struct value *v) {
         break;
     case V_TRANSFORM:
         fprintf(out, "<transform:");
-        print_info(stdout, v->transform->lens->info);
+        print_info(out, v->transform->lens->info);
         fprintf(out, ">");
         break;
     case V_NATIVE:
         fprintf(out, "<native:");
-        print_info(stdout, v->info);
+        print_info(out, v->info);
         fprintf(out, ">");
         break;
     case V_CLOS:
         fprintf(out, "<closure:");
-        print_info(stdout, v->func->info);
+        print_info(out, v->func->info);
         fprintf(out, ">");
         break;
     case V_EXN:
-        print_info(stdout, v->exn->info);
-        fprintf(out, "exception: %s\n", v->exn->message);
-        for (int i=0; i < v->exn->nlines; i++) {
-            fprintf(out, "    %s\n", v->exn->lines[i]);
+        if (! v->exn->seen) {
+            print_info(out, v->exn->info);
+            fprintf(out, "exception: %s\n", v->exn->message);
+            for (int i=0; i < v->exn->nlines; i++) {
+                fprintf(out, "    %s\n", v->exn->lines[i]);
+            }
+            v->exn->seen = 1;
         }
         break;
     default:
@@ -928,8 +944,8 @@ static struct value *coerce(struct value *v, struct type *t) {
         unref(vt, type);
         return rxp;
     }
-    fatal_error(v->info, "Failed to coerce %s to %s",
-                type_name(vt), type_name(t));
+    return make_exn_value(v->info, "Type %s can not be coerced to %s",
+                          type_name(vt), type_name(t));
 }
 
 /* Return one of the expected types (passed as ...).
@@ -1131,9 +1147,13 @@ static int check_binop(const char *opname, struct term *term,
 }
 
 static int check_value(struct value *v) {
+    const char *msg;
+
     if (v->tag == V_REGEXP) {
-        if (regexp_compile(v->regexp) == -1)
+        if (regexp_check(v->regexp, &msg) == -1) {
+            syntax_error(v->info, "Invalid regular expression: %s", msg);
             return 0;
+        }
     }
     return 1;
 }
@@ -1142,7 +1162,7 @@ static int check_value(struct value *v) {
 static int check_exp(struct term *term, struct ctx *ctx) {
     int result = 1;
     assert(term->type == NULL || term->tag == A_VALUE || term->ref > 1);
-    if (term->type != NULL)
+    if (term->type != NULL && term->tag != A_VALUE)
         return 1;
 
     switch (term->tag) {
@@ -1347,10 +1367,17 @@ static struct value *compile_union(struct term *exp, struct ctx *ctx) {
 
     struct type *t = exp->type;
     struct info *info = exp->info;
-    struct value *v;
+    struct value *v = NULL;
 
     v1 = coerce(v1, t);
+    if (EXN(v1))
+        return v1;
     v2 = coerce(v2, t);
+    if (EXN(v2)) {
+        unref(v1, value);
+        return v2;
+    }
+
     if (t->tag == T_REGEXP) {
         v = make_value(V_REGEXP, ref(info));
         v->regexp = regexp_union(info, v1->regexp, v2->regexp);
@@ -1642,25 +1669,6 @@ static struct value *compile_exp(struct info *info,
     return v;
 }
 
-static void print_tree(FILE *out, int indent, struct tree *tree) {
-    list_for_each(t, tree) {
-        for (int i=0; i < indent; i++) fputc(' ', out);
-        fprintf(out, "{ ");
-        if (t->label != NULL)
-            fprintf(out, "\"%s\"", t->label);
-        if (t->value != NULL)
-            fprintf(out, " = \"%s\"", t->value);
-        if (t->children != NULL) {
-            fputc('\n', out);
-            print_tree(out, indent + 2, t->children);
-            for (int i=0; i < indent; i++) fputc(' ', out);
-        } else {
-            fputc(' ', out);
-        }
-        fprintf(out, "}\n");
-    }
-}
-
 static int compile_test(struct term *term, struct ctx *ctx) {
     struct value *actual = compile_exp(term->info, term->test, ctx);
     struct value *expect = NULL;
@@ -1675,7 +1683,8 @@ static int compile_test(struct term *term, struct ctx *ctx) {
         }
     } else {
         if (EXN(actual)) {
-            printf("Test run encountered exception:\n");
+            print_info(stdout, term->info);
+            printf("exception thrown in test\n");
             print_value(stdout, actual);
             printf("\n");
             ret = 0;
@@ -1721,11 +1730,20 @@ static int compile_decl(struct term *term, struct ctx *ctx) {
         bind(&ctx->local, term->bname, term->type, v);
 
         if (EXN(v) && !v->exn->seen) {
+            struct error *error = term->info->error;
+            struct memstream ms;
+
+            init_memstream(&ms);
+
             syntax_error(term->info, "Failed to compile %s",
                          term->bname);
-            print_value(stdout, v);
-            printf("\n");
+            fprintf(ms.stream, "%s\n", error->details);
+            print_value(ms.stream, v);
+            close_memstream(&ms);
+
             v->exn->seen = 1;
+            free(error->details);
+            error->details = ms.buf;
         }
         result = ! EXN(v);
         unref(v, value);
@@ -1740,17 +1758,15 @@ static int compile_decl(struct term *term, struct ctx *ctx) {
 static struct module *compile(struct term *term, struct augeas *aug) {
     struct ctx ctx;
     struct transform *autoload = NULL;
-    int ok = 1;
     assert(term->tag == A_MODULE);
 
     ctx.aug = aug;
     ctx.local = NULL;
     ctx.name = term->mname;
     list_for_each(dcl, term->decls) {
-        ok &= compile_decl(dcl, &ctx);
+        if (!compile_decl(dcl, &ctx))
+            goto error;
     }
-    if (!ok)
-        goto error;
 
     if (term->autoload != NULL) {
         struct binding *bnd = bnd_lookup(ctx.local, term->autoload);
@@ -1775,27 +1791,40 @@ static struct module *compile(struct term *term, struct augeas *aug) {
 /*
  * Defining native functions
  */
-static struct info *make_native_info(const char *fname, int line) {
+static struct info *
+make_native_info(struct error *error, const char *fname, int line) {
     struct info *info;
-    make_ref(info);
+    if (make_ref(info) < 0)
+        goto error;
     info->first_line = info->last_line = line;
     info->first_column = info->last_column = 0;
-    make_ref(info->filename);
+    info->error = error;
+    if (make_ref(info->filename) < 0)
+        goto error;
     info->filename->str = strdup(fname);
     return info;
+ error:
+    unref(info, info);
+    return NULL;
 }
 
-void define_native_intl(const char *file, int line,
-                        struct module *module, const char *name,
-                        int argc, void *impl, ...) {
+int define_native_intl(const char *file, int line,
+                       struct error *error,
+                       struct module *module, const char *name,
+                       int argc, void *impl, ...) {
     assert(argc > 0);  /* We have no unit type */
     assert(argc <= 5);
     va_list ap;
     enum type_tag tag;
-    struct term *params = NULL, *body;
+    struct term *params = NULL, *body = NULL, *func = NULL;
     struct type *type;
-    struct value *v;
-    struct info *info = make_native_info(file, line);
+    struct value *v = NULL;
+    struct info *info = NULL;
+    struct ctx ctx;
+
+    info = make_native_info(error, file, line);
+    if (info == NULL)
+        goto error;
 
     va_start(ap, impl);
     for (int i=0; i < argc; i++) {
@@ -1813,22 +1842,32 @@ void define_native_intl(const char *file, int line,
     type = make_base_type(tag);
 
     make_ref(v);
+    if (v == NULL)
+        goto error;
     v->tag = V_NATIVE;
     v->info = info;
+    info = NULL;
 
-    CALLOC(v->native, 1);
+    if (ALLOC(v->native) < 0)
+        goto error;
     v->native->argc = argc;
     v->native->type = type;
     v->native->impl = impl;
 
     make_ref(body);
+    if (body == NULL)
+        goto error;
     body->info = ref(info);
     body->type = ref(type);
     body->tag = A_VALUE;
     body->value = v;
+    v = NULL;
+
+    func = build_func(params, body);
+    if (func == NULL)
+        goto error;
+    body = NULL;
 
-    struct term *func = build_func(params, body);
-    struct ctx ctx;
     ctx.aug = NULL;
     ctx.local = ref(module->bindings);
     ctx.name = module->name;
@@ -1838,17 +1877,27 @@ void define_native_intl(const char *file, int line,
         abort();
     }
     v = make_closure(func, ctx.local);
+    if (v == NULL) {
+        unref(module->bindings, binding);
+        goto error;
+    }
     bind(&ctx.local, name, func->type, v);
     unref(v, value);
     unref(func, term);
     unref(module->bindings, binding);
 
     module->bindings = ctx.local;
+    return 0;
+ error:
+    unref(v, value);
+    unref(body, term);
+    unref(func, term);
+    return -1;
 }
 
 
 /* Defined in parser.y */
-int augl_parse_file(const char *name, struct term **term);
+int augl_parse_file(struct augeas *aug, const char *name, struct term **term);
 
 static char *module_basename(const char *modname) {
     char *fname;
@@ -1882,12 +1931,12 @@ static char *module_filename(struct augeas *aug, const char *modname) {
     return filename;
 }
 
-int __aug_load_module_file(struct augeas *aug, const char *filename) {
+int load_module_file(struct augeas *aug, const char *filename) {
     struct term *term = NULL;
     int result = -1;
 
-    if (augl_parse_file(filename, &term) == -1)
-        goto error;
+    augl_parse_file(aug, filename, &term);
+    ERR_BAIL(aug);
 
     if (! typecheck(term, aug))
         goto error;
@@ -1914,7 +1963,7 @@ static int load_module(struct augeas *aug, const char *name) {
     if ((filename = module_filename(aug, name)) == NULL)
         return -1;
 
-    if (__aug_load_module_file(aug, filename) == -1)
+    if (load_module_file(aug, filename) == -1)
         goto error;
 
     free(filename);
@@ -1928,7 +1977,7 @@ static int load_module(struct augeas *aug, const char *name) {
 int interpreter_init(struct augeas *aug) {
     int r;
 
-    aug->modules = builtin_init();
+    aug->modules = builtin_init(aug->error);
 
     if (aug->flags & AUG_NO_MODL_AUTOLOAD)
         return 0;
@@ -1943,11 +1992,18 @@ int interpreter_init(struct augeas *aug) {
     while ((dir = argz_next(aug->modpathz, aug->nmodpath, dir)) != NULL) {
         char *globpat;
         r = asprintf(&globpat, "%s/*.aug", dir);
-        if (r == -1)
-            return -1;
+        ERR_NOMEM(r < 0, aug);
+
         r = glob(globpat, gl_flags, NULL, &globbuf);
-        if (r != 0 && r != GLOB_NOMATCH)
-            FIXME("glob failure for %s", globpat);
+        if (r != 0 && r != GLOB_NOMATCH) {
+            /* This really has to be an allocation failure; glob is not
+             * supposed to return GLOB_ABORTED here */
+            aug_errcode_t code =
+                r == GLOB_NOSPACE ? AUG_ENOMEM : AUG_EINTERNAL;
+            ERR_REPORT(aug, code, "glob failure for %s", globpat);
+            free(globpat);
+            goto error;
+        }
         gl_flags |= GLOB_APPEND;
         free(globpat);
     }
@@ -1969,7 +2025,7 @@ int interpreter_init(struct augeas *aug) {
     globfree(&globbuf);
     return 0;
  error:
-    FIXME("Cleanup loaded modules");
+    globfree(&globbuf);
     return -1;
 }
 
