@@ -22,8 +22,6 @@
 
 #include <config.h>
 
-#include "canonicalize.h"
-
 #include <fnmatch.h>
 #include <glob.h>
 
@@ -38,6 +36,7 @@
 #include "augeas.h"
 #include "syntax.h"
 #include "transform.h"
+#include "errcode.h"
 
 static const int fnm_flags = FNM_PATHNAME;
 static const int glob_flags = GLOB_NOSORT;
@@ -61,7 +60,6 @@ static const int glob_flags = GLOB_NOSORT;
 static const char *const s_path = "path";
 static const char *const s_lens = "lens";
 static const char *const s_info = "info";
-static const char *const s_id   = "id";
 
 static const char *const s_error = "error";
 /* These are all put underneath "error" */
@@ -103,6 +101,16 @@ static bool is_incl(struct tree *f) {
     return streqv(f->label, "incl") && f->value != NULL;
 }
 
+static bool is_regular_file(const char *path) {
+    int r;
+    struct stat st;
+
+    r = stat(path, &st);
+    if (r < 0)
+        return false;
+    return S_ISREG(st.st_mode);
+}
+
 static int filter_generate(struct tree *xfm, const char *root,
                            int *nmatches, char ***matches) {
     glob_t globbuf;
@@ -124,44 +132,59 @@ static int filter_generate(struct tree *xfm, const char *root,
 
         if (r != 0 && r != GLOB_NOMATCH) {
             ret = -1;
-            goto done;
+            goto error;
         }
         gl_flags |= GLOB_APPEND;
     }
 
-    char **pathv = globbuf.gl_pathv;
-    int pathc = globbuf.gl_pathc;
-    globbuf.gl_pathv = NULL;
-    globbuf.gl_pathc = 0;
+    char **pathv = NULL;
+    int pathc = globbuf.gl_pathc, pathind = 0;
+
+    if (ALLOC_N(pathv, pathc) < 0)
+        goto error;
+
+    for (int i=0; i < pathc; i++) {
+        const char *path = globbuf.gl_pathv[i];
+        bool include = true;
+
+        list_for_each(e, xfm->children) {
+            if (! is_excl(e))
+                continue;
 
-    list_for_each(e, xfm->children) {
-        if (! is_excl(e))
-            continue;
-        for (int i=0; i < pathc;) {
-            const char *path = pathv[i];
             if (strchr(e->value, SEP) == NULL)
                 path = pathbase(path);
-            if (fnmatch(e->value, path, fnm_flags) == 0) {
-                free(pathv[i]);
-                pathc -= 1;
-                if (i < pathc) {
-                    pathv[i] = pathv[pathc];
-                }
-            } else {
-                i += 1;
+            if ((r = fnmatch(e->value, path, fnm_flags)) == 0) {
+                include = false;
             }
         }
+
+        if (include)
+            include = is_regular_file(globbuf.gl_pathv[i]);
+
+        if (include) {
+            pathv[pathind] = strdup(globbuf.gl_pathv[i]);
+            if (pathv[pathind] == NULL)
+                goto error;
+            pathind += 1;
+        }
     }
-    if (REALLOC_N(pathv, pathc) == -1) {
-        FREE(pathv);
-        pathc = 0;
-        ret = -1;
-    }
+    pathc = pathind;
+
+    if (REALLOC_N(pathv, pathc) == -1)
+        goto error;
+
     *matches = pathv;
     *nmatches = pathc;
  done:
     globfree(&globbuf);
     return ret;
+ error:
+    if (pathv != NULL)
+        for (int i=0; i < pathc; i++)
+            free(pathv[i]);
+    free(pathv);
+    ret = -1;
+    goto done;
 }
 
 static int filter_matches(struct tree *xfm, const char *path) {
@@ -237,20 +260,6 @@ static int err_set(struct augeas *aug, char **ep, const char *sub,
     return (r < 0) ? -1 : 0;
 }
 
-static void
-calc_line_ofs(const char *text, size_t pos, size_t *line, size_t *ofs)
-{
-    *line = 1;
-    *ofs = 0;
-    for (const char *t = text; t < text + pos; t++) {
-        *ofs += 1;
-        if (*t == '\n') {
-            *ofs = 0;
-            *line += 1;
-        }
-    }
-}
-
 /* Record an error in the tree. The error will show up underneath
  * /augeas/FILENAME/error. PATH is the path to the toplevel node in the
  * tree where the lens application happened. When STATUS is NULL, just
@@ -308,56 +317,55 @@ static int store_error(struct augeas *aug,
     return result;
 }
 
-static int add_load_info(struct augeas *aug, const char *filename,
-                         const char *node, struct lens *lens) {
+/* Set up the file information in the /augeas tree.
+ *
+ * NODE must be the path to the file contents, and start with /files.
+ * LENS is the lens used to transform the file.
+ * Create entries under /augeas/NODE with some metadata about the file.
+ *
+ * Returns 0 on success, -1 on error
+ */
+static int add_file_info(struct augeas *aug, const char *node,
+                         struct lens *lens, const char *lens_name) {
+    struct tree *file, *tree;
     char *tmp = NULL;
     int r;
-    char *p = NULL;
-    int end = 0;
+    char *path = NULL;
     int result = -1;
 
-    r = pathjoin(&p, 2, AUGEAS_META_FILES, filename + strlen(aug->root) - 1);
-    if (r < 0)
-        goto done;
-    end = strlen(p);
+    if (lens == NULL)
+        return -1;
 
-    r = pathjoin(&p, 1, s_path);
-    if (r < 0)
-        goto done;
+    r = pathjoin(&path, 2, AUGEAS_META_TREE, node);
+    ERR_NOMEM(r < 0, aug);
 
-    r = aug_set(aug, p, node);
-    if (r < 0)
-        goto done;
-    p[end] = '\0';
+    file = tree_find_cr(aug, path);
+    ERR_BAIL(aug);
 
-    r = pathjoin(&p, 2, s_lens, s_info);
-    if (r < 0)
-        goto done;
+    /* Set 'path' */
+    tree = tree_child_cr(file, s_path);
+    ERR_NOMEM(tree == NULL, aug);
+    r = tree_set_value(tree, node);
+    ERR_NOMEM(r < 0, aug);
 
+    /* Set 'lens/info' */
     tmp = format_info(lens->info);
-    if (tmp == NULL)
-        goto done;
-    r = aug_set(aug, p, tmp);
+    ERR_NOMEM(tmp == NULL, aug);
+    tree = tree_path_cr(file, 2, s_lens, s_info);
+    ERR_NOMEM(tree == NULL, aug);
+    r = tree_set_value(tree, tmp);
+    ERR_NOMEM(r < 0, aug);
     FREE(tmp);
-    if (r < 0)
-        goto done;
-    p[end] = '\0';
 
-    r = pathjoin(&p, 2, s_lens, s_id);
-    if (r < 0)
-        goto done;
-
-    r = asprintf(&tmp, "%p", lens);
-    if (r >= 0) {
-        r = aug_set(aug, p, tmp);
-        FREE(tmp);
-        if (r < 0)
-            goto done;
-    }
+    /* Set 'lens' */
+    tree = tree->parent;
+    r = tree_set_value(tree, lens_name);
+    ERR_NOMEM(r < 0, aug);
 
     result = 0;
- done:
-    free(p);
+ error:
+    free(path);
+    free(tmp);
     return result;
 }
 
@@ -374,7 +382,17 @@ static char *append_newline(char *text, size_t len) {
     return text;
 }
 
-static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
+/* Turn the file name FNAME, which starts with aug->root, into
+ * a path in the tree underneath /files */
+static char *file_name_path(struct augeas *aug, const char *fname) {
+    char *path = NULL;
+
+    pathjoin(&path, 2, AUGEAS_FILES_TREE, fname + strlen(aug->root) - 1);
+    return path;
+}
+
+static int load_file(struct augeas *aug, struct lens *lens,
+                     const char *lens_name, char *filename) {
     char *text = NULL;
     const char *err_status = NULL;
     struct aug_file *file = NULL;
@@ -383,13 +401,14 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
     struct lns_error *err = NULL;
     int result = -1, r;
 
-    pathjoin(&path, 2, AUGEAS_FILES_TREE, filename + strlen(aug->root) - 1);
+    path = file_name_path(aug, filename);
+    ERR_NOMEM(path == NULL, aug);
 
-    r = add_load_info(aug, filename, path, lens);
+    r = add_file_info(aug, path, lens, lens_name);
     if (r < 0)
         goto done;
 
-    text = read_file(filename);
+    text = xread_file(filename);
     if (text == NULL) {
         err_status = "read_failed";
         goto done;
@@ -400,6 +419,7 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
     make_ref(info);
     make_ref(info->filename);
     info->filename->str = filename;
+    info->error = aug->error;
     info->first_line = 1;
 
     tree = lns_get(info, lens, text, &err);
@@ -419,6 +439,7 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
  done:
     store_error(aug, filename + strlen(aug->root) - 1, path, err_status,
                 errno, err, text);
+ error:
     free_lns_error(err);
     free(path);
     free_tree(tree);
@@ -433,20 +454,40 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
  * autoload transform for Module
  */
 static struct lens *lens_from_name(struct augeas *aug, const char *name) {
+    struct lens *result = NULL;
+
     if (name[0] == '@') {
         struct module *modl = NULL;
         for (modl = aug->modules;
              modl != NULL && !streqv(modl->name, name + 1);
              modl = modl->next);
-        if (modl == NULL || modl->autoload == NULL)
-            return NULL;
-        return modl->autoload->lens;
+        ERR_THROW(modl == NULL, aug, AUG_ENOLENS,
+                  "Could not find module %s", name + 1);
+        ERR_THROW(modl->autoload == NULL, aug, AUG_ENOLENS,
+                  "No autoloaded lens in module %s", name + 1);
+        result = modl->autoload->lens;
     } else {
-        return lens_lookup(aug, name);
+        result = lens_lookup(aug, name);
     }
+    ERR_THROW(result == NULL, aug, AUG_ENOLENS,
+              "Can not find lens %s", name);
+    return result;
+ error:
+    return NULL;
 }
 
-static struct lens *xfm_lens(struct augeas *aug, struct tree *xfm) {
+const char *xfm_lens_name(struct tree *xfm) {
+    struct tree *l = tree_child(xfm, s_lens);
+
+    if (l == NULL)
+        return "(unknown)";
+    if (l->value == NULL)
+        return "(noname)";
+    return l->value;
+}
+
+static struct lens *xfm_lens(struct augeas *aug,
+                             struct tree *xfm, const char **lens_name) {
     struct tree *l = NULL;
 
     for (l = xfm->children;
@@ -455,6 +496,8 @@ static struct lens *xfm_lens(struct augeas *aug, struct tree *xfm) {
 
     if (l == NULL || l->value == NULL)
         return NULL;
+    *lens_name = l->value;
+
     return lens_from_name(aug, l->value);
 }
 
@@ -491,23 +534,61 @@ int transform_validate(struct augeas *aug, struct tree *xfm) {
         xfm_error(xfm, "the 'lens' node does not contain a lens name");
         return -1;
     }
-    if (lens_from_name(aug, l->value) == NULL) {
-        char *msg;
-        if (asprintf(&msg, "the lens '%s' does not exist", l->value) < 0) {
-            xfm_error(xfm, "the lens does not exist");
-        } else {
-            xfm_error(xfm, msg);
-            free(msg);
-        }
+    lens_from_name(aug, l->value);
+    if (HAS_ERR(aug)) {
+        xfm_error(xfm, aug->error->details);
         return -1;
     }
     return 0;
 }
 
+void transform_file_error(struct augeas *aug, const char *status,
+                          const char *filename, const char *format, ...) {
+    char *ep = err_path(filename);
+    struct tree *err;
+    char *msg;
+    va_list ap;
+    int r;
+
+    err = tree_find_cr(aug, ep);
+    if (err == NULL)
+        return;
+
+    tree_unlink_children(aug, err);
+    tree_set_value(err, status);
+
+    err = tree_child_cr(err, s_message);
+    if (err == NULL)
+        return;
+
+    va_start(ap, format);
+    r = vasprintf(&msg, format, ap);
+    va_end(ap);
+    if (r < 0)
+        return;
+    tree_set_value(err, msg);
+    free(msg);
+}
+
+static struct tree *file_info(struct augeas *aug, const char *fname) {
+    char *path = NULL;
+    struct tree *result = NULL;
+    int r;
+
+    r = pathjoin(&path, 2, AUGEAS_META_FILES, fname);
+    ERR_NOMEM(r < 0, aug);
+
+    result = tree_find(aug, path);
+    ERR_BAIL(aug);
+ error:
+    return result;
+}
+
 int transform_load(struct augeas *aug, struct tree *xfm) {
-    int nmatches;
+    int nmatches = 0;
     char **matches;
-    struct lens *lens = xfm_lens(aug, xfm);
+    const char *lens_name;
+    struct lens *lens = xfm_lens(aug, xfm, &lens_name);
     int r;
 
     if (lens == NULL) {
@@ -518,8 +599,20 @@ int transform_load(struct augeas *aug, struct tree *xfm) {
     if (r == -1)
         return -1;
     for (int i=0; i < nmatches; i++) {
-        load_file(aug, lens, matches[i]);
-        free(matches[i]);
+        const char *filename = matches[i] + strlen(aug->root) - 1;
+        struct tree *finfo = file_info(aug, filename);
+        if (finfo != NULL && tree_child(finfo, s_lens) != NULL) {
+            const char *s = xfm_lens_name(finfo);
+            char *fpath = file_name_path(aug, matches[i]);
+            transform_file_error(aug, "mxfm_load", filename,
+              "Lenses %s and %s could be used to load this file",
+                                 s, lens_name);
+            aug_rm(aug, fpath);
+            free(fpath);
+        } else {
+            load_file(aug, lens, lens_name, matches[i]);
+        }
+        FREE(matches[i]);
     }
     lens_release(lens);
     free(matches);
@@ -655,7 +748,8 @@ static int file_saved_event(struct augeas *aug, const char *path) {
     struct tree *dummy;
     int r;
 
-    r = pathx_parse(aug->origin, AUGEAS_EVENTS_SAVED "[last()]",
+    r = pathx_parse(aug->origin, err_of_aug(aug),
+                    AUGEAS_EVENTS_SAVED "[last()]",
                     true, NULL, &px);
     if (r != PATHX_NOERROR)
         return -1;
@@ -701,7 +795,8 @@ int transform_save(struct augeas *aug, struct tree *xfm,
     const char *err_status = NULL;
     char *dyn_err_status = NULL;
     struct lns_error *err = NULL;
-    struct lens *lens = xfm_lens(aug, xfm);
+    const char *lens_name;
+    struct lens *lens = xfm_lens(aug, xfm, &lens_name);
     int result = -1, r;
 
     errno = 0;
@@ -719,13 +814,8 @@ int transform_save(struct augeas *aug, struct tree *xfm,
         goto done;
     }
 
-    if (asprintf(&augnew, "%s%s" EXT_AUGNEW, aug->root, filename) == -1) {
-        augnew = NULL;
-        goto done;
-    }
-
     if (access(augorig, R_OK) == 0) {
-        text = read_file(augorig);
+        text = xread_file(augorig);
     } else {
         text = strdup("");
     }
@@ -737,15 +827,6 @@ int transform_save(struct augeas *aug, struct tree *xfm,
 
     text = append_newline(text, strlen(text));
 
-    // FIXME: We might have to create intermediary directories
-    // to be able to write augnew, but we have no idea what permissions
-    // etc. they should get. Just the process default ?
-    fp = fopen(augnew, "w");
-    if (fp == NULL) {
-        err_status = "open_augnew";
-        goto done;
-    }
-
     augorig_canon = canonicalize_file_name(augorig);
     augorig_exists = 1;
     if (augorig_canon == NULL) {
@@ -757,6 +838,30 @@ int transform_save(struct augeas *aug, struct tree *xfm,
             goto done;
         }
     }
+
+    /* Figure out where to put the .augnew file. If we need to rename it
+       later on, put it next to augorig_canon */
+    if (aug->flags & AUG_SAVE_NEWFILE) {
+        if (xasprintf(&augnew, "%s" EXT_AUGNEW, augorig) < 0) {
+            err_status = "augnew_oom";
+            goto done;
+        }
+    } else {
+        if (xasprintf(&augnew, "%s" EXT_AUGNEW, augorig_canon) < 0) {
+            err_status = "augnew_oom";
+            goto done;
+        }
+    }
+
+    // FIXME: We might have to create intermediate directories
+    // to be able to write augnew, but we have no idea what permissions
+    // etc. they should get. Just the process default ?
+    fp = fopen(augnew, "w");
+    if (fp == NULL) {
+        err_status = "open_augnew";
+        goto done;
+    }
+
     if (augorig_exists) {
         if (transfer_file_attrs(augorig_canon, augnew, &err_status) != 0) {
             err_status = "xfer_attrs";
@@ -796,7 +901,7 @@ int transform_save(struct augeas *aug, struct tree *xfm,
     }
 
     {
-        char *new_text = read_file(augnew);
+        char *new_text = xread_file(augnew);
         int same = 0;
         if (new_text == NULL) {
             err_status = "read_augnew";
@@ -839,6 +944,11 @@ int transform_save(struct augeas *aug, struct tree *xfm,
     result = 1;
 
  done:
+    r = add_file_info(aug, path, lens, lens_name);
+    if (r < 0) {
+        err_status = "file_info";
+        result = -1;
+    }
     if (result > 0) {
         r = file_saved_event(aug, path);
         if (r < 0) {
