@@ -25,8 +25,10 @@
 #include "fa.h"
 #include "cutest.h"
 #include "internal.h"
+#include "memory.h"
 #include <stdio.h>
 #include <stdlib.h>
+#include <ctype.h>
 
 #define FA_DOT_DIR "FA_DOT_DIR"
 
@@ -43,7 +45,13 @@ static void print_regerror(int err, const char *regexp) {
     size = regerror(err, NULL, NULL, 0);
     errbuf = alloca(size);
     regerror(err, NULL, errbuf, size);
-    fprintf(stderr, "Error building fa from %s:\n", regexp);
+    if (strlen(regexp) > 40) {
+        char *s = strndup(regexp, 40);
+        fprintf(stderr, "Error building fa from %s...:\n", s);
+        free(s);
+    } else {
+        fprintf(stderr, "Error building fa from %s:\n", regexp);
+    }
     fprintf(stderr, "  %s\n", errbuf);
 }
 
@@ -62,7 +70,8 @@ static struct fa *mark(struct fa *fa) {
     struct fa_list *fl;
 
     if (fa != NULL) {
-        CALLOC(fl, 1);
+        if (ALLOC(fl) < 0)
+            die_oom();
         fl->fa = fa;
         list_cons(fa_list, fl);
     }
@@ -80,25 +89,33 @@ static void assertAsRegexp(CuTest *tc, struct fa *fa) {
     fa1 = mark(fa_concat(fa, empty));
     /* Minimize FA1, otherwise the regexp returned is enormous for the */
     /* monster (~ 2MB) and fa_compile becomes incredibly slow          */
-    fa_minimize(fa1);
+    r = fa_minimize(fa1);
+    if (r < 0)
+        die_oom();
 
     r = fa_as_regexp(fa1, &re, &re_len);
     CuAssertIntEquals(tc, 0, r);
 
-    r = fa_compile(re, strlen(re), &fa2);
+    r = fa_compile(re, re_len, &fa2);
+    if (r != REG_NOERROR) {
+        print_regerror(r, re);
+    }
     CuAssertIntEquals(tc, REG_NOERROR, r);
-
     CuAssertTrue(tc, fa_equals(fa, fa2));
 
     fa_free(fa2);
     free(re);
 }
 
-static struct fa *make_fa(CuTest *tc, const char *regexp, int exp_err) {
+static struct fa *make_fa(CuTest *tc,
+                          const char *regexp, size_t reglen,
+                          int exp_err) {
     struct fa *fa;
     int r;
 
-    r = fa_compile(regexp, strlen(regexp), &fa);
+    r = fa_compile(regexp, reglen, &fa);
+    if (r == REG_ESPACE)
+        die_oom();
     if (exp_err == REG_NOERROR) {
         if (r != REG_NOERROR)
             print_regerror(r, regexp);
@@ -114,7 +131,7 @@ static struct fa *make_fa(CuTest *tc, const char *regexp, int exp_err) {
 }
 
 static struct fa *make_good_fa(CuTest *tc, const char *regexp) {
-    return make_fa(tc, regexp, REG_NOERROR);
+    return make_fa(tc, regexp, strlen(regexp), REG_NOERROR);
 }
 
 static void dot(struct fa *fa) {
@@ -143,8 +160,10 @@ static void dot(struct fa *fa) {
 }
 
 static void testBadRegexps(CuTest *tc) {
-    make_fa(tc, "(x", REG_EPAREN);
-    make_fa(tc, "a{5,3}", REG_BADBR);
+    const char *const re1 = "(x";
+    const char *const re2 = "a{5,3}";
+    make_fa(tc, re1, strlen(re1), REG_EPAREN);
+    make_fa(tc, re2, strlen(re2), REG_BADBR);
 }
 
 /* Stress test, mostly good to check that allocation is clean */
@@ -311,6 +330,15 @@ static void assertExample(CuTest *tc, const char *regexp, const char *exp) {
     fa_example(fa, &xmpl, &xmpl_len);
     CuAssertStrEquals(tc, exp, xmpl);
     free(xmpl);
+
+    fa_nocase(fa);
+    char *s = strdup(exp);
+    for (int i = 0; i < strlen(s); i++) s[i] = tolower(s[i]);
+
+    fa_example(fa, &xmpl, &xmpl_len);
+    CuAssertStrEquals(tc, s, xmpl);
+    free(xmpl);
+    free(s);
 }
 
 static void testExample(CuTest *tc) {
@@ -331,6 +359,8 @@ static void testExample(CuTest *tc) {
     assertExample(tc, "\001((\002.)*\001)+\002", "\001\001\002");
     assertExample(tc, "\001((\001.)*\002)+\002", "\001\002\002");
 
+    assertExample(tc, "a[^\001-\004]+b", "aAb");
+
     struct fa *fa1 = mark(fa_make_basic(FA_EMPTY));
     size_t xmpl_len;
     char *xmpl;
@@ -410,6 +440,8 @@ static void testAsRegexp(CuTest *tc) {
     assertFaAsRegexp(tc, "ab|cd");
     assertFaAsRegexp(tc, "[a-z]+");
     assertFaAsRegexp(tc, "[]a-]+");
+    assertFaAsRegexp(tc, "[^0-9A-Z]");
+    assertFaAsRegexp(tc, "ab|(xy[A-Z0-9])*(uv[^0-9]?)");
     assertFaAsRegexp(tc, "[A-CE-GI-LN-QS-Z]");
 }
 
@@ -431,7 +463,103 @@ static void testAsRegexpMinus(CuTest *tc) {
 }
 
 static void testRangeEnd(CuTest *tc) {
-    make_fa(tc, "[1-0]", REG_ERANGE);
+    const char *const re = "[1-0]";
+    make_fa(tc, re, strlen(re), REG_ERANGE);
+}
+
+static void testNul(CuTest *tc) {
+    static const char *const re0 = "a\0b";
+    int re0_len = 3;
+
+    struct fa *fa1 = make_fa(tc, "a\0b", re0_len, REG_NOERROR);
+    struct fa *fa2 = make_good_fa(tc, "a.b");
+    char *re;
+    size_t re_len;
+    int r;
+
+    CuAssertTrue(tc, fa_contains(fa1, fa2));
+
+    r = fa_as_regexp(fa1, &re, &re_len);
+    CuAssertIntEquals(tc, 0, r);
+    CuAssertIntEquals(tc, re0_len, re_len);
+    CuAssertIntEquals(tc, 0, memcmp(re0, re, re0_len));
+}
+
+static void testRestrictAlphabet(CuTest *tc) {
+    const char *re = "ab|(xy[B-Z0-9])*(uv[^0-9]?)";
+    struct fa *fa_exp = make_good_fa(tc, "((xy[0-9])*)uv[^0-9A-Z]?|ab");
+    struct fa *fa_act = NULL;
+    size_t nre_len;
+    char *nre;
+    int r;
+
+    r = fa_restrict_alphabet(re, strlen(re), &nre, &nre_len, 'A', 'Z');
+    CuAssertIntEquals(tc, 0, r);
+    CuAssertIntEquals(tc, nre_len, strlen(nre));
+    fa_act = make_good_fa(tc, nre);
+    CuAssertTrue(tc, fa_equals(fa_exp, fa_act));
+    free(nre);
+
+    r = fa_restrict_alphabet("HELLO", strlen("HELLO"),
+                             &nre, &nre_len, 'A', 'Z');
+    CuAssertIntEquals(tc, -2, r);
+    CuAssertPtrEquals(tc, NULL, nre);
+
+    r = fa_restrict_alphabet("a{2,", strlen("a{2"), &nre, &nre_len, 'A', 'Z');
+    CuAssertIntEquals(tc, REG_EBRACE, r);
+}
+
+static void testExpandCharRanges(CuTest *tc) {
+    const char *re = "[1-3]*|[a-b]([^\nU-X][^\n])*";
+    char *nre;
+    size_t nre_len;
+    int r;
+
+    r = fa_expand_char_ranges(re, strlen(re), &nre, &nre_len);
+    CuAssertIntEquals(tc, 0, r);
+    CuAssertStrEquals(tc, "[123]*|[ab]([^\nUVWX].)*", nre);
+    CuAssertIntEquals(tc, strlen(nre), nre_len);
+}
+
+static void testNoCase(CuTest *tc) {
+    struct fa *fa1 = make_good_fa(tc, "[a-z0-9]");
+    struct fa *fa2 = make_good_fa(tc, "B");
+    struct fa *fa, *exp;
+    int r;
+
+    fa_nocase(fa1);
+    fa = fa_intersect(fa1, fa2);
+    CuAssertPtrNotNull(tc, fa);
+    r = fa_equals(fa, fa2);
+    fa_free(fa);
+    CuAssertIntEquals(tc, 1, r);
+
+    fa = fa_concat(fa1, fa2);
+    exp = make_good_fa(tc, "[a-zA-Z0-9]B");
+    r = fa_equals(fa, exp);
+    fa_free(fa);
+    CuAssertIntEquals(tc, 1, r);
+}
+
+static void testExpandNoCase(CuTest *tc) {
+    const char *p1 = "aB";
+    const char *p2 = "[a-cUV]";
+    const char *p3 = "[^a-z]";
+    char *s;
+    size_t len;
+    int r;
+
+    r = fa_expand_nocase(p1, strlen(p1), &s, &len);
+    CuAssertStrEquals(tc, "[Aa][Bb]", s);
+    free(s);
+
+    r = fa_expand_nocase(p2, strlen(p2), &s, &len);
+    CuAssertStrEquals(tc, "[A-CUVa-cuv]", s);
+    free(s);
+
+    r = fa_expand_nocase(p3, strlen(p3), &s, &len);
+    CuAssertStrEquals(tc, "[^A-Za-z]", s);
+    free(s);
 }
 
 int main(int argc, char **argv) {
@@ -453,13 +581,21 @@ int main(int argc, char **argv) {
         SUITE_ADD_TEST(suite, testAsRegexp);
         SUITE_ADD_TEST(suite, testAsRegexpMinus);
         SUITE_ADD_TEST(suite, testRangeEnd);
+        SUITE_ADD_TEST(suite, testNul);
+        SUITE_ADD_TEST(suite, testRestrictAlphabet);
+        SUITE_ADD_TEST(suite, testExpandCharRanges);
+        SUITE_ADD_TEST(suite, testNoCase);
+        SUITE_ADD_TEST(suite, testExpandNoCase);
 
         CuSuiteRun(suite);
         CuSuiteSummary(suite, &output);
         CuSuiteDetails(suite, &output);
         printf("%s\n", output);
         free(output);
-        return suite->failCount;
+        if (getenv("FAILMALLOC_INTERVAL") != NULL)
+            return EXIT_SUCCESS;
+        else
+            return suite->failCount;
     }
 
     for (int i=1; i<argc; i++) {
