@@ -18,7 +18,7 @@ myfile {
 EOF
 ln $file $file.orig
 
-augtool -I $abs_top_srcdir/lenses -r $root > /dev/null <<EOF
+augtool --nostdinc -I $abs_top_srcdir/lenses -r $root > /dev/null <<EOF
 ins invalid before /files/etc/logrotate.d/test/rule
 save
 EOF
