@@ -158,6 +158,58 @@ static void testInvalidLens(CuTest *tc) {
     aug_close(aug);
 }
 
+static void testLoadSave(CuTest *tc) {
+    augeas *aug = NULL;
+    int r;
+
+    /* FIXME: This test behaves properly during distcheck, since srcdir
+     * is writeprotected, making an incorrect attempt to write
+     * /etc/hosts.augnew fail; during normal 'make check' the test will
+     * succeed.
+     * To address this, we should copy the files fro, tests/root into
+     * another directory and 'chmod a-w /etc' in that root
+     */
+    aug = aug_init(root, loadpath, AUG_NO_MODL_AUTOLOAD|AUG_SAVE_NOOP);
+    CuAssertPtrNotNull(tc, aug);
+
+    r = aug_set(aug, "/augeas/load/Hosts/lens", "Hosts.lns");
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_set(aug, "/augeas/load/Hosts/incl", "/etc/hosts");
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_load(aug);
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_save(aug);
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_match(aug, "/augeas/events/saved", NULL);
+    CuAssertIntEquals(tc, 0, r);
+
+    aug_close(aug);
+}
+
+/* Tests bug #79 */
+static void testLoadDefined(CuTest *tc) {
+    augeas *aug = NULL;
+    int r;
+
+    aug = aug_init(root, loadpath, AUG_NO_STDINC);
+    CuAssertPtrNotNull(tc, aug);
+
+    r = aug_defvar(aug, "v", "/files/etc/hosts/*/ipaddr");
+    CuAssertIntEquals(tc, 2, r);
+
+    r = aug_load(aug);
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_match(aug, "$v", NULL);
+    CuAssertIntEquals(tc, 0, r);
+
+    aug_close(aug);
+}
+
 int main(void) {
     char *output = NULL;
     CuSuite* suite = CuSuiteNew();
@@ -167,6 +219,8 @@ int main(void) {
     SUITE_ADD_TEST(suite, testNoLoad);
     SUITE_ADD_TEST(suite, testNoAutoload);
     SUITE_ADD_TEST(suite, testInvalidLens);
+    SUITE_ADD_TEST(suite, testLoadSave);
+    SUITE_ADD_TEST(suite, testLoadDefined);
 
     abs_top_srcdir = getenv("abs_top_srcdir");
     if (abs_top_srcdir == NULL)
